"""The module contains the implementation of the EntityConsumerMdib."""
from __future__ import annotations

import copy
import time
from dataclasses import dataclass
from threading import Lock, Thread
from typing import TYPE_CHECKING, Any, Callable, Protocol, cast

from sdc11073 import loghelper
from sdc11073 import observableproperties as properties
from sdc11073.exceptions import ApiUsageError
from sdc11073.mdib.consumermdib import ConsumerMdibState
from sdc11073.mdib.mdibbase import MdibVersionGroup
from sdc11073.namespaces import QN_TYPE, default_ns_helper
from sdc11073.xml_types import msg_qnames, pm_qnames

from .entities import ConsumerMultiStateEntity, XmlEntity, XmlMultiStateEntity, get_xsi_type
from .entity_consumermdibxtra import EntityConsumerMdibMethods
from .entity_mdibbase import EntityMdibBase

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lxml.etree import QName

    from sdc11073.consumer.consumerimpl import SdcConsumer
    from sdc11073.mdib.entityprotocol import EntityGetterProtocol
    from sdc11073.pysoap.msgreader import ReceivedMessage
    from sdc11073.xml_types.pm_types import CodedValue, Coding
    from sdc11073.xml_utils import LxmlElement

    from .entities import ConsumerEntityType, ConsumerInternalEntityType

    XmlEntityFactory = Callable[[LxmlElement, str, str], ConsumerInternalEntityType]


@dataclass
class _BufferedData:
    received_message_data: ReceivedMessage
    handler: Callable


multi_state_q_names = (pm_qnames.PatientContextDescriptor,
                       pm_qnames.LocationContextDescriptor,
                       pm_qnames.WorkflowContextDescriptor,
                       pm_qnames.OperatorContextDescriptor,
                       pm_qnames.MeansContextDescriptor,
                       pm_qnames.EnsembleContextDescriptor)


def _mk_xml_entity(node: LxmlElement, parent_handle: str, source_mds: str) -> ConsumerInternalEntityType:
    """Return a new XmlEntity or XmlMultiStateEntity.

    This is the default consumer entity factory.
    It creates one of ConsumerInternalEntityType, which are factories for ConsumerEntityType.
    By using a different factory, the user can change this to use other classes.
    """
    xsi_type = get_xsi_type(node)
    if xsi_type in multi_state_q_names:
        return XmlMultiStateEntity(parent_handle, source_mds, xsi_type, node, [])
    return XmlEntity(parent_handle, source_mds, xsi_type, node, None)


class EntityGetter:
    """Implementation of EntityGetterProtocol."""

    def __init__(self, entities: dict[str, XmlEntity | XmlMultiStateEntity], mdib: EntityConsumerMdib):
        self._entities = entities
        self._mdib = mdib

    def by_handle(self, handle: str) -> ConsumerEntityType | None:
        """Return entity with given descriptor handle."""
        try:
            return self._mk_entity(handle)
        except KeyError:
            return None

    def by_context_handle(self, handle: str) -> ConsumerMultiStateEntity | None:
        """Return multi state entity that contains a state with given handle."""
        for internal_entity in self._entities.values():
            if internal_entity.is_multi_state and handle in internal_entity.states:
                _internal_entity = cast(XmlMultiStateEntity, internal_entity)
                return _internal_entity.mk_entity(self._mdib)
        return None

    def by_node_type(self, node_type: QName) -> list[ConsumerEntityType]:
        """Return all entities with given node type."""
        return [self._mk_entity(handle) for handle, entity in self._entities.items()
                if entity.node_type == node_type]

    def by_parent_handle(self, parent_handle: str | None) -> list[ConsumerEntityType]:
        """Return all entities with given parent handle."""
        return [self._mk_entity(handle) for handle, entity in self._entities.items()
                if entity.parent_handle == parent_handle]

    def by_coding(self, coding: Coding) -> list[ConsumerEntityType]:
        """Return all entities with given Coding."""
        return [self._mk_entity(handle) for handle, entity in self._entities.items()
                if entity.coded_value is not None and entity.coded_value.is_equivalent(coding)]

    def by_coded_value(self, coded_value: CodedValue) -> list[ConsumerEntityType]:
        """Return all entities with given Coding."""
        return [self._mk_entity(handle) for handle, entity in self._entities.items()
                if entity.coded_value is not None and entity.coded_value.is_equivalent(coded_value)]

    def items(self) -> Iterable[tuple[str, ConsumerEntityType]]:
        """Return items of a dictionary."""
        for handle in self._entities:
            yield handle, self._mk_entity(handle)

    def _mk_entity(self, handle: str) -> ConsumerEntityType:
        xml_entity = self._mdib.internal_entities[handle]
        return xml_entity.mk_entity(self._mdib)

    def __len__(self) -> int:
        """Return number of entities."""
        return len(self._entities)

class ConsumerXtraProtocol(Protocol):  # pragma: no cover
    """Functionality expected by EntityConsumerMdib."""

    def bind_to_client_observables(self):
        """Connect mdib to consumer."""
        ...


class EntityConsumerMdib(EntityMdibBase):
    """Implementation of the consumer side mdib with EntityGetter Interface.

    The internal entities store descriptors and states as XML nodes. This needs only very little CPU time for
    handling of notifications.
    The instantiation of descriptor and state container instances is only done on demand when the user calls the
    EntityGetter interface.
    """

    sequence_or_instance_id_changed_event: bool = properties.ObservableProperty(
        default_value=False, fire_only_on_changed_value=False)
    # sequence_or_instance_id_changed_event is set to True every time the sequence id changes.
    # It is not reset to False any time later.
    # It is in the responsibility of the application to react on a changed sequence id.
    # Observe this property and call "reload_all" in the observer code.

    MDIB_VERSION_CHECK_DISABLED = False
    # for testing purpose you can disable checking of mdib version, so that every notification is accepted.

    def __init__(self,
                 sdc_client: SdcConsumer,
                 extras_cls: Callable[[EntityConsumerMdib, loghelper.LoggerAdapter],
                 ConsumerXtraProtocol] | None = None,
                 max_realtime_samples: int = 100,
                 maintain_xml_tree: bool = False):
        """Construct a ConsumerMdib instance.

        :param sdc_client: a SdcConsumer instance
        :param  extras_cls: extended functionality
        :param max_realtime_samples: determines how many real time samples are stored per RealtimeSampleArray
        :param maintain_xml_tree: Enable or disable the maintenance of a xml tree in self.get_mdib_response_node.
               If set, the initial GetMdibResponse data is saved there and updated when notifications are received.
        """
        super().__init__(sdc_client.sdc_definitions,
                         loghelper.get_logger_adapter('sdc.client.mdib', sdc_client.log_prefix))
        self._entity_factory: XmlEntityFactory = _mk_xml_entity
        self._entities: dict[str, XmlEntity | XmlMultiStateEntity] = {}  # key is the handle

        self._sdc_client = sdc_client
        if extras_cls is None:
            extras_cls = EntityConsumerMdibMethods
        self._xtra = extras_cls(self, self._logger)
        self._state = ConsumerMdibState.invalid
        self.rt_buffers = {}  # key  is a handle, value is a ConsumerRtBuffer
        self._max_realtime_samples = max_realtime_samples
        self._maintain_xml_tree = maintain_xml_tree
        self._last_wf_age_log = time.time()
        # a buffer for notifications that are received before initial get_mdib is done
        self._buffered_notifications = []
        self._buffered_notifications_lock = Lock()
        self.entities: EntityGetterProtocol = EntityGetter(self._entities, self)

        self.get_mdib_response_node: LxmlElement | None = None
        self._md_description_node: LxmlElement | None = None
        self._md_state_node: LxmlElement | None = None

    @property
    def xtra(self) -> Any:
        """Give access to extended functionality."""
        return self._xtra

    @property
    def internal_entities(self) -> dict[str, ConsumerInternalEntityType]:
        """The property is needed by transactions. Do not use it otherwise."""
        return self._entities

    @property
    def sdc_client(self) -> SdcConsumer:
        """Give access to sdc client."""
        return self._sdc_client

    @property
    def is_initialized(self) -> bool:
        """Returns True if everything has been set up completely."""
        return self._state == ConsumerMdibState.initialized

    def init_mdib(self):
        """Binds own notification handlers to observables of sdc client and calls GetMdib.

        Client mdib is initialized from GetMdibResponse, and from then on updated from incoming notifications.
        :return: None
        """
        if self.is_initialized: # pragma: no cover
            raise ApiUsageError('ConsumerMdib is already initialized')
        # first start receiving notifications, then call get_mdib.
        # Otherwise, we might miss notifications.
        self._xtra.bind_to_client_observables()
        self.reload_all()
        self._sdc_client.set_mdib(self)
        self._logger.info('initializing mdib done')

    def reload_all(self):
        """Delete all data and reloads everything."""
        self._logger.info('reload_all called')
        with self.mdib_lock:
            self._state = ConsumerMdibState.initializing  # notifications are now buffered
            self.get_mdib_response_node = None
            self._md_state_node = None  # this is the parent node of all states
            self.sequence_id = None
            self.instance_id = None
            self.mdib_version = None

            get_service = self._sdc_client.client('Get')
            self._logger.info('initializing mdib...')
            response = get_service.get_mdib()
            self._set_root_node(response.p_msg.msg_node)
            self._update_mdib_version_group(cast(MdibVersionGroup, response.mdib_version_group))

            # process buffered notifications
            with self._buffered_notifications_lock:
                self._logger.debug('got _buffered_notifications_lock')
                for buffered_report in self._buffered_notifications:
                    # buffered data might contain notifications that do not fit.
                    mvg = buffered_report.received_message_data.mdib_version_group
                    if mvg.sequence_id != self.sequence_id:
                        self.logger.debug('wrong sequence id "%s"; ignore buffered report',
                                          mvg.sequence_id)
                        continue
                    if mvg.mdib_version <= self.mdib_version:
                        self.logger.debug('older mdib version "%d"; ignore buffered report',
                                          mvg.mdib_version)
                        continue
                    buffered_report.handler(buffered_report.received_message_data)
                del self._buffered_notifications[:]
                self._state = ConsumerMdibState.initialized
            self._logger.info('reload_all done')

    def process_incoming_metric_states_report(self, received_message_data: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message_data,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_metric_states_report(received_message_data)

    def _process_incoming_metric_states_report(self, received_message: ReceivedMessage):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        self.metric_handles = self._process_incoming_state_report(received_message,
                                                                  msg_qnames.MetricState)

    def process_incoming_alert_states_report(self, received_message_data: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message_data,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_alert_states_report(received_message_data)

    def _process_incoming_alert_states_report(self, received_message: ReceivedMessage):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        self.alert_handles = self._process_incoming_state_report(received_message,
                                                                 msg_qnames.AlertState)

    def process_incoming_component_report(self, received_message_data: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message_data,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_component_report(received_message_data)

    def _process_incoming_component_report(self, received_message: ReceivedMessage):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        self.component_handles = self._process_incoming_state_report(received_message,
                                                                     msg_qnames.ComponentState)

    def process_incoming_operational_state_report(self, received_message_data: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message_data,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_operational_state_report(received_message_data)

    def _process_incoming_operational_state_report(self, received_message: ReceivedMessage):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        self.operation_handles = self._process_incoming_state_report(received_message,
                                                                     msg_qnames.OperationState)

    def process_incoming_context_report(self, received_message_data: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message_data,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_context_report(received_message_data)

    def _process_incoming_context_report(self, received_message: ReceivedMessage):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        if self._can_accept_mdib_version(received_message.mdib_version_group.mdib_version, 'component states'):
            self._update_mdib_version_group(cast(MdibVersionGroup, received_message.mdib_version_group))

        handles = []
        # msg_node can be None
        report_parts = received_message.p_msg.msg_node if received_message.p_msg.msg_node is not None else []
        for report_part in report_parts:
            for state_node in report_part:
                if state_node.tag == msg_qnames.ContextState:
                    handle = state_node.attrib['Handle']
                    descriptor_handle = state_node.attrib['DescriptorHandle']
                    xml_entity = self._entities[descriptor_handle]
                    # modify state_node, but only in a deep copy
                    state_node = copy.deepcopy(state_node)  # noqa: PLW2901
                    state_node.tag = pm_qnames.State  # keep old tag

                    if self._maintain_xml_tree:
                        # replace old xml state in self._md_state_node with new one (keep xml tree up to date)
                        found = False
                        for old_st in self._md_state_node:
                            if old_st.get('Handle') == handle:
                                self._md_state_node.replace(old_st, state_node)
                                found = True
                                break
                        if not found:
                            self._md_state_node.append(state_node)

                    # replace or add in xml entity
                    xml_entity.states[handle] = state_node

                    handles.append(handle)
        self.context_handles = handles  # update observable

    def process_incoming_waveform_states(self, received_message_data: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message_data,
                                         self._process_incoming_waveform_states):
            return
        with self.mdib_lock:
            self._process_incoming_waveform_states(received_message_data)

    def _process_incoming_waveform_states(self, received_message: ReceivedMessage):
        if self._can_accept_mdib_version(received_message.mdib_version_group.mdib_version, 'waveform states'):
            self._update_mdib_version_group(cast(MdibVersionGroup, received_message.mdib_version_group))

        handles = []
        # msg_node can be None
        states = received_message.p_msg.msg_node if received_message.p_msg.msg_node is not None else []
        for state in states:
            # _update_state replaces states in entity and optionally in self.get_mdib_response_node
            handles.append(self._update_state(state, pm_qnames.RealTimeSampleArrayMetricState)) # noqa: PERF401
        self.waveform_handles = handles  # update observable

    def process_incoming_description_modification_report(self, received_message: ReceivedMessage):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(received_message,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_description_modification_report(received_message)

    def _process_incoming_description_modification_report(self, received_message: ReceivedMessage):
        new_descriptors_handles = []
        updated_descriptors_handles = []
        deleted_descriptors_handles = []
        # msg_node can be None
        report_parts = received_message.p_msg.msg_node if received_message.p_msg.msg_node is not None else []
        for report_part in report_parts:
            parent_handle = report_part.attrib.get('ParentDescriptor')  # can be None in case of MDS, but that is ok
            source_mds_handle = report_part[0].text
            descriptors = [copy.deepcopy(e) for e in report_part if e.tag == msg_qnames.Descriptor]
            states = [copy.deepcopy(e) for e in report_part if e.tag == msg_qnames.State]
            modification_type = report_part.attrib.get('ModificationType', 'Upt')  # implied value is 'Upt'
            if modification_type == 'Upt':
                updated_descriptors_handles.extend(self._update_descriptors(parent_handle,
                                                                            source_mds_handle,
                                                                            descriptors,
                                                                            states))
            elif modification_type == 'Crt':
                new_descriptors_handles.extend(self._create_descriptors(parent_handle,
                                                                        source_mds_handle,
                                                                        descriptors,
                                                                        states))
            elif modification_type == 'Del':
                deleted_descriptors_handles.extend(self._delete_descriptors(descriptors))
            else:
                self.logger.error('Unknown modification type %r', modification_type)
        if updated_descriptors_handles:
            self.updated_descriptors_handles = updated_descriptors_handles
        if new_descriptors_handles:
            self.new_descriptors_handles = new_descriptors_handles
        if deleted_descriptors_handles:
            self.deleted_descriptors_handles = deleted_descriptors_handles

    def _update_descriptors(self,
                            parent_handle: str,
                            source_mds_handle: str,
                            descriptors: list[LxmlElement],
                            states: list[LxmlElement]) -> list[str]:
        handles = []
        for descriptor in descriptors:
            handle = descriptor.attrib['Handle']
            entity = self._entities.get(handle)
            if entity is None:
                self.logger.error('got descriptor update for not existing handle "%s"', handle)
                continue
            if entity.parent_handle != parent_handle:
                self.logger.error('inconsistent parent handle "%s" for "%s"', handle, entity.parent_handle)
                continue
            if entity.source_mds != source_mds_handle:
                self.logger.error('inconsistent source mds handle "%s" for "%s"',
                                  source_mds_handle, entity.source_mds)
                continue
            current_states = [s for s in states if s.attrib['DescriptorHandle'] == handle]
            self._update_descriptor_states(descriptor, current_states)
            handles.append(handle)
        return handles

    def _create_descriptors(self, # noqa: PLR0912 C901
                            parent_handle: str,
                            source_mds_handle: str,
                            descriptors: list[LxmlElement],
                            states: list[LxmlElement]) -> list[str]:
        handles = []
        for descriptor in descriptors:
            descriptor_handle = descriptor.attrib['Handle']
            current_states = [s for s in states if s.attrib['DescriptorHandle'] == descriptor_handle]

            if self._maintain_xml_tree:
                # add states to parent (MdState node)
                for st in current_states:
                    st.tag = pm_qnames.State
                    self._md_state_node.append(st)

            xml_entity = self._entity_factory(descriptor, parent_handle, source_mds_handle)
            if xml_entity.is_multi_state:
                for st in current_states:
                    xml_entity.states[st.attrib['Handle']] = st
            elif len(current_states) != 1:
                self.logger.error('create descriptor: Expect one state, got %d', len(current_states))
            else:
                xml_entity.state = current_states[0]
            self._entities[descriptor_handle] = xml_entity
            handles.append(descriptor_handle)

            # add descriptor to parent
            parent_xml_entity = self._entities[parent_handle]
            if parent_xml_entity.node_type == pm_qnames.ChannelDescriptor:
                # channel children have same tag
                descriptor.tag = pm_qnames.Metric
                if self._maintain_xml_tree:
                    parent_xml_entity.descriptor.append(descriptor)
            elif parent_xml_entity.node_type == pm_qnames.VmdDescriptor:
                # vmd children have same tag
                descriptor.tag = pm_qnames.Channel
                if self._maintain_xml_tree:
                    parent_xml_entity.descriptor.append(descriptor)
            elif parent_xml_entity.node_type == pm_qnames.MdsDescriptor:
                # Mds children have different names.
                # child_order determines the tag of the element (first tuple position), and the corresponding type
                # (2nd position)
                xsi_type = get_xsi_type(descriptor)
                # the list serves 2 tasks:
                # - define the tag name based on xsi type
                # - define the order of children, if xml tree is maintained
                child_order: list[tuple[QName, QName]] = [
                    (pm_qnames.MetaData, pm_qnames.MetaData),  # optional member, no handle
                    (pm_qnames.SystemContext, pm_qnames.SystemContextDescriptor),
                    (pm_qnames.Clock, pm_qnames.ClockDescriptor),
                    (pm_qnames.Battery, pm_qnames.BatteryDescriptor),
                    (pm_qnames.ApprovedJurisdictions, pm_qnames.ApprovedJurisdictions),
                    # optional list, no handle
                    (pm_qnames.Vmd, pm_qnames.VmdDescriptor)]
                # rename node, based on xsi_type
                for entry in child_order:
                    if xsi_type == entry[1]:
                        descriptor.tag = entry[0]
                        break

                if self._maintain_xml_tree:
                    # Insert at correct position
                    self._insert_child(descriptor,
                                       parent_xml_entity.descriptor,
                                       [ch[0] for ch in child_order])
        return handles

    def _insert_child(self,
                      child_node: LxmlElement,
                      parent_node: LxmlElement,
                      child_order: list[QName]):
        """Insert at correct position."""
        add_before_q_names = []

        for i, schema_name in enumerate(child_order):
            if schema_name == child_node.tag:
                add_before_q_names.extend(child_order[i + 1:])
                break

        # find position
        existing_children = parent_node[:]
        for tmp_child_node in existing_children:
            if tmp_child_node.tag in add_before_q_names:
                tmp_child_node.addprevious(child_node)
                return
        parent_node.append(child_node)

    def _delete_descriptors(self, descriptors: list[LxmlElement]) -> list[str]:
        handles = []
        for descriptor in descriptors:
            handle = descriptor.attrib['Handle']
            entity = self._entities.get(handle)
            if entity is None:
                self.logger.error('shall delete descriptor "%s", but it is unknown', handle)
            else:
                self._delete_entity(entity, handles)
        return handles

    def _delete_entity(self, entity: XmlEntity | XmlMultiStateEntity, deleted_handles: list[str]):
        """Recursive method to delete an entity and subtree."""
        parent = entity.descriptor.getparent()
        if parent is not None:
            parent.remove(entity.descriptor)
        if entity.is_multi_state:
            states = entity.states.values()
        elif entity.state is None:
            states = []
        else:
            states = [entity.state]
        for state in states:
            parent = state.getparent()
            if parent is not None:
                parent.remove(state)
        handle = entity.descriptor.get('Handle')
        del self._entities[handle]
        deleted_handles.append(handle)
        child_entities = [e for e in self._entities.values() if e.parent_handle == handle]
        for e in child_entities:
            self._delete_entity(e, deleted_handles)

    def _process_incoming_state_report(self, received_message: ReceivedMessage, expected_q_name: QName) -> list[str]:
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        if self._can_accept_mdib_version(received_message.mdib_version_group.mdib_version, 'state'):
            self._update_mdib_version_group(cast(MdibVersionGroup, received_message.mdib_version_group))

        handles = []
        report_parts = received_message.p_msg.msg_node if received_message.p_msg.msg_node is not None else []
        for report_part in report_parts:
            for state in report_part:
                if state.tag == expected_q_name:
                    # _update_state replaces states in entity and optionally in self.get_mdib_response_node
                    handles.append(self._update_state(state)) # noqa: PERF401
        return handles  # update observable

    def _update_state(self, state_node: LxmlElement, xsi_type: QName | None = None) -> str:
        """Replace state in entity and DOM tree."""
        descriptor_handle = state_node.attrib['DescriptorHandle']
        xml_entity = self._entities[descriptor_handle]
        state_node = copy.deepcopy(state_node)  # we modify state_node, but only in a deep copy. Copy has no parent!
        state_node.tag = pm_qnames.State  # xml_entity.state.tag  # keep old tag
        if xsi_type:
            state_node.set(QN_TYPE, default_ns_helper.doc_name_from_qname(xsi_type))

        if self._maintain_xml_tree:
            # replace old xml state in parent_node with new one (keep xml tree up to date)
            if xml_entity.state is not None:
                parent = xml_entity.state.getparent()
                parent.replace(xml_entity.state, state_node)
        # replace in xml entity
        xml_entity.state = state_node
        return descriptor_handle

    def _update_descriptor_states(self, # noqa: PLR0912 C901
                                  descriptor_node: LxmlElement,
                                  state_nodes: list[LxmlElement]) -> str:
        """Replace state in DOM tree and entity."""
        for state_node in state_nodes:
            state_node.tag = pm_qnames.State  # rename in order to have a valid tag acc. to participant model

        descriptor_handle = descriptor_node.attrib['Handle']
        xml_entity = self._entities[descriptor_handle]

        if not xml_entity.is_multi_state:
            if len(state_nodes) == 0:
                self.logger.error('Update descriptor %s: no state provided. State will not be updated.',
                                  descriptor_handle)
            elif len(state_nodes) > 1:
                self.logger.error('Update descriptor %s: expected 1 state, got %d. Will use only 1st one.',
                                  descriptor_handle, len(state_nodes))

        descriptor_node.tag = xml_entity.descriptor.tag  # keep old tag

        if self._maintain_xml_tree:
            # replace descriptor in xml tree with the new one:
            # 1. move all children of xml_entity.descriptor with a Handle
            # 2. from entity.descriptor to descriptor_node (at identical position )
            children = xml_entity.descriptor[:]
            for idx, child in enumerate(children):
                if 'Handle' in child.attrib:
                    descriptor_node.insert(idx, child)
            # replace descriptor in parent
            descriptor_parent = xml_entity.descriptor.getparent()
            descriptor_parent.replace(xml_entity.descriptor, descriptor_node)

        # replace descriptor in xml_entity
        xml_entity.descriptor = descriptor_node

        if self._maintain_xml_tree:
            if xml_entity.is_multi_state:
                current_handles = []
                # replace state_nodes in self._md_state_node
                for state_node in state_nodes:
                    st_handle = state_node.attrib['Handle']
                    current_handles.append(st_handle)
                    old_state = xml_entity.states.get(st_handle)
                    if old_state is not None:
                        self._md_state_node.replace(old_state, state_node)
                    else:
                        self._md_state_node.append(state_node)
                # delete states in self._md_state_node that are not in current list of state nodes
                for handle, state in xml_entity.states.items():
                    if handle not in current_handles:
                        self._md_state_node.remove(state)
            elif len(state_nodes) >= 1:
                if xml_entity.state is not None :
                    self._md_state_node.replace(xml_entity.state, state_nodes[0])
                else:
                    self._md_state_node.append(state_nodes[0])
        if xml_entity.is_multi_state:
            # replace state_nodes in xml_entity
            xml_entity.states.clear()
            for state_node in state_nodes:
                xml_entity.states[state_node.attrib['Handle']] = state_node
        elif len(state_nodes) >= 1:
            # replace state_node in xml_entity
            xml_entity.state = state_nodes[0]
        return descriptor_handle

    def _pre_check_report_ok(self, received_message_data: ReceivedMessage,
                             handler: Callable) -> bool:
        """Check if the report can be added to mdib.

        The pre-check runs before the mdib lock is acquired.
        The report is buffered if state is 'initializing' and 'is_buffered_report' is False.
        :return: True if report can be added to mdib.
        """
        self._check_sequence_or_instance_id_changed(
            received_message_data.mdib_version_group)  # this might change self._state
        if self._state == ConsumerMdibState.invalid:
            # ignore report in these states
            return False
        if self._state == ConsumerMdibState.initializing:
            with self._buffered_notifications_lock:
                # check state again, it might have changed before lock was acquired
                if self._state == ConsumerMdibState.initializing:
                    self._buffered_notifications.append(_BufferedData(received_message_data, handler))
                    return False
        return True

    def _can_accept_mdib_version(self, new_mdib_version: int, log_prefix: str) -> bool:
        if self.MDIB_VERSION_CHECK_DISABLED:  # pragma: no cover
            return True
        # log deviations from expected mdib version
        if new_mdib_version < self.mdib_version:
            self._logger.warning('{}: ignoring too old Mdib version, have {}, got {}',  # noqa: PLE1205
                                 log_prefix, self.mdib_version, new_mdib_version)
        elif (new_mdib_version - self.mdib_version) > 1:
            # This can happen if consumer did not subscribe to all notifications.
            # Still log a warning, because mdib is no longer a correct mirror of provider mdib.
            self._logger.warning('{}: expect mdib_version {}, got {}',  # noqa: PLE1205
                                 log_prefix, self.mdib_version + 1, new_mdib_version)
        # it is possible to receive multiple notifications with the same mdib version => compare ">="
        return new_mdib_version >= self.mdib_version

    def _check_sequence_or_instance_id_changed(self, mdib_version_group: MdibVersionGroup):
        """Check if sequence id and instance id are still the same.

        If not,
        - set state member to invalid
        - set the observable "sequence_or_instance_id_changed_event" in a thread.
          This allows to implement an observer that can directly call reload_all without blocking the consumer.
        """
        if mdib_version_group.sequence_id == self.sequence_id and mdib_version_group.instance_id == self.instance_id:
            return
        if self._state == ConsumerMdibState.initialized:
            if mdib_version_group.sequence_id != self.sequence_id:
                self.logger.warning('sequence id changed from "%s" to "%s"',
                                    self.sequence_id, mdib_version_group.sequence_id)
            if mdib_version_group.instance_id != self.instance_id:
                self.logger.warning('instance id changed from "%r" to "%r"',
                                    self.instance_id, mdib_version_group.instance_id)
            self.logger.warning('mdib is no longer valid!')

            self._state = ConsumerMdibState.invalid

            def _set_observable():
                self.sequence_or_instance_id_changed_event = True

            thr = Thread(target=_set_observable)
            thr.start()

    def _set_root_node(self, root_node: LxmlElement): # noqa: C901
        """Set member and create xml entities."""
        if root_node.tag != msg_qnames.GetMdibResponse:  # pragma: no cover
            msg = f'root node must be {msg_qnames.GetMdibResponse!s}, got {root_node.tag!s}'
            raise ValueError(msg)
        self._entities.clear()
        md_state_node = None
        md_description_node = None

        # look for md_state_node and md_description_node
        for child_element in root_node[0]:  # Mdib node has children MdDescription, MdState; both are optional
            if child_element.tag == pm_qnames.MdState:
                md_state_node = child_element
            if child_element.tag == pm_qnames.MdDescription:
                md_description_node = child_element

        def register_children_with_handle(parent_node: LxmlElement, source_mds: str | None = None):
            parent_handle = parent_node.attrib.get('Handle')
            for child_node in parent_node[:]:
                child_handle = child_node.attrib.get('Handle')
                if child_node.tag == pm_qnames.Mds:
                    source_mds = child_handle
                if child_handle:
                    self._entities[child_handle] = self._entity_factory(child_node,
                                                                        parent_handle,
                                                                        source_mds)
                    register_children_with_handle(child_node, source_mds)

        if md_description_node is not None:
            register_children_with_handle(md_description_node)

        if md_state_node is not None:
            for state_node in md_state_node:
                descriptor_handle = state_node.attrib['DescriptorHandle']
                entity = self._entities[descriptor_handle]
                if entity.is_multi_state:
                    handle = state_node.attrib['Handle']
                    entity.states[handle] = state_node
                else:
                    entity.state = state_node

        if self._maintain_xml_tree:
            # set self.get_mdib_response_node, self._md_state_node and self._md_description_node
            self.get_mdib_response_node = root_node
            self._md_state_node = md_state_node
            self._md_description_node = md_description_node


    def _update_mdib_version_group(self, mdib_version_group: MdibVersionGroup):
        """Set members and optionally update entries in DOM tree."""
        if mdib_version_group.mdib_version != self.mdib_version:
            self.mdib_version = mdib_version_group.mdib_version
            if self._maintain_xml_tree and self.get_mdib_response_node is not None:
                self.get_mdib_response_node.set('MdibVersion', str(mdib_version_group.mdib_version))
                self.get_mdib_response_node[0].set('MdibVersion', str(mdib_version_group.mdib_version))
        if mdib_version_group.sequence_id != self.sequence_id:
            self.sequence_id = mdib_version_group.sequence_id
            if self._maintain_xml_tree and self.get_mdib_response_node is not None:
                self.get_mdib_response_node.set('SequenceId', str(mdib_version_group.sequence_id))
                self.get_mdib_response_node[0].set('SequenceId', str(mdib_version_group.sequence_id))
        if mdib_version_group.instance_id != self.instance_id:
            self.instance_id = mdib_version_group.instance_id
            if self. _maintain_xml_tree and self.get_mdib_response_node is not None:
                self.get_mdib_response_node.set('InstanceId', str(mdib_version_group.instance_id))
                self.get_mdib_response_node[0].set('InstanceId', str(mdib_version_group.instance_id))




"""The module implements the base class of role providers."""
from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from lxml import etree

from sdc11073 import loghelper
from sdc11073.provider.operations import OperationDefinitionBase

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorProtocol, AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.mdib.transactionsprotocol import AbstractTransactionManagerProtocol
    from sdc11073.provider.operations import ExecuteHandler, TimeoutHandler
    from sdc11073.provider.sco import AbstractScoOperationsRegistry
    from sdc11073.xml_types.pm_types import CodedValue, SafetyClassification

OperationClassGetter = Callable[[etree.QName], type[OperationDefinitionBase]]


class ProviderRole:
    """Base class for all role implementations."""

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str):
        self._mdib = mdib
        self._logger = loghelper.get_logger_adapter(f'sdc.device.{self.__class__.__name__}', log_prefix)

    def stop(self):
        """Stop whatever needs to be stopped.

        Implement method in derived class if needed.
        """

    def init_operations(self, sco: AbstractScoOperationsRegistry):
        """Initialize and start whatever the provider needs.

        Implement method in derived class if needed.
        """

    def make_operation_instance(
            self,
            operation_descriptor_container: AbstractOperationDescriptorProtocol,  # noqa: ARG002
            operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:  # noqa: ARG002
        """Return an operation definition instance for this operation or None.

        If a mdib already has operations defined, this method can connect a handler to a given operation descriptor.
        Use case: initialization from an existing mdib.
        """
        return None

    def make_missing_operations(self, sco: AbstractScoOperationsRegistry) -> list[  # noqa: ARG002
        OperationDefinitionBase]:
        """Create operations that this role provider needs.

        This method is called after all existing operations from mdib have been registered.
        If a role provider needs to add operations beyond that, it can do it here.
        """
        return []

    def on_pre_commit(self, mdib: ProviderMdibProtocol, transaction: AbstractTransactionManagerProtocol):
        """Manipulate the transaction if needed.

        Derived classes can overwrite this method.
        """

    def on_post_commit(self, mdib: ProviderMdibProtocol, transaction: AbstractTransactionManagerProtocol):
        """Run stuff after transaction.

        Derived classes can overwrite this method.
        """

    def _mk_operation_from_operation_descriptor(self,
                                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                                operation_cls_getter: OperationClassGetter,
                                                operation_handler: ExecuteHandler,
                                                timeout_handler: TimeoutHandler | None = None) \
            -> OperationDefinitionBase:
        """Create an OperationDefinition for the operation_descriptor_container."""
        op_cls = operation_cls_getter(operation_descriptor_container.NODETYPE)
        return op_cls(operation_descriptor_container.Handle,
                      operation_descriptor_container.OperationTarget,
                      operation_handler=operation_handler,
                      timeout_handler=timeout_handler,
                      coded_value=operation_descriptor_container.Type)

    @staticmethod
    def _create_descriptor_container(container_cls: type[AbstractDescriptorProtocol],
                                     handle: str,
                                     parent_handle: str,
                                     coded_value: CodedValue,
                                     safety_classification: SafetyClassification) -> AbstractDescriptorProtocol:
        obj = container_cls(handle=handle, parent_handle=parent_handle)
        obj.SafetyClassification = safety_classification
        obj.Type = coded_value
        return obj



from __future__ import annotations

from http.server import BaseHTTPRequestHandler
from urllib.parse import urlparse

from .compression import CompressionHandler
from .httpreader import HTTPReader, mk_chunks
from sdc11073.exceptions import InvalidPathError


class DispatchingRequestHandler(BaseHTTPRequestHandler):
    """This request handler expects that the http server has a 'dispatcher' member that is a PathElementRegistry.
     The registered Components must be a MessageConverterMiddleware. """
    protocol_version = "HTTP/1.1"  # this enables keep-alive

    # This server does NOT disable nagle algorithm. It sends Large responses,
    # and network efficiency is more important than short latencies.
    disable_nagle_algorithm = False

    def _read_request(self):
        """ checks header for content-length, chunk-encoding and compression entries.
        Handles incoming bytes correspondingly.
        :return: http body as bytes
        """
        return HTTPReader.read_request_body(self)

    def _compress_if_supported(self, response_bytes: bytes):
        """Compress response if header of request indicates that other side
        accepts one of our supported compression encodings"""
        accepted_enc = CompressionHandler.parse_header(self.headers.get('accept-encoding'))
        for enc in accepted_enc:
            if enc in self.server.supported_encodings:
                response_bytes = CompressionHandler.compress_payload(enc, response_bytes)
                self.send_header('Content-Encoding', enc)
                break
        return response_bytes

    def log_request(self, code='-', size='-'):
        pass  # suppress printing of every request to stderr

    def get_first_path_element(self):
        parsed_path = urlparse(self.path)
        path_elements = parsed_path.path.split('/')
        if len(path_elements[0]) > 0:
            return path_elements[0]
        return path_elements[1]

    def do_POST(self):  # pylint: disable=invalid-name
        request_bytes = self._read_request()
        if self.server.dispatcher is None:
            # close this connection
            self.close_connection = True  # pylint: disable=attribute-defined-outside-init
            http_reason = 'received a POST request, but have no dispatcher'
            response_xml_string = http_reason.encode('utf-8')
            self.send_response(500, http_reason)  # server error
            self.send_header("Content-type", "text/plain; charset=utf-8")
            self.send_header("Content-length", str(len(response_xml_string)))
            self.end_headers()
            self.wfile.write(response_xml_string)
            return
        try:
            component = self.server.dispatcher.get_instance(self.get_first_path_element())
        except InvalidPathError as ex:
            self.server.logger.error('invalid path {} (request from {}): {}', self.path, self.client_address, ex.reason)
            http_reason = ex.reason
            response_xml_string = b''
            self.send_response(ex.status, http_reason)
            self.send_header("Content-type", "text/plain; charset=utf-8")
            self.send_header("Content-length", str(len(response_xml_string)))
            self.end_headers()
            self.wfile.write(response_xml_string)
            return

        peer_name = self.connection.getpeername()
        try:
            result = component.do_post(self.headers, self.path, peer_name, request_bytes)
            http_status, http_reason, response_xml_string = result
        except Exception as ex:
            self.server.logger.error('exception (request from {}): {}', self.path, self.client_address, ex)
            http_reason = str(ex)
            response_xml_string = b''
            self.send_response(500, http_reason)  # server error
            self.send_header("Content-type", "text/plain; charset=utf-8")
            self.send_header("Content-length", str(len(response_xml_string)))
            self.end_headers()
            self.wfile.write(response_xml_string)
            return

        self.send_response(http_status, http_reason)
        response_xml_string = self._compress_if_supported(response_xml_string)
        self.send_header("Content-type", "application/soap+xml; charset=utf-8")
        if self.server.chunk_size > 0:
            self.send_header("transfer-encoding", "chunked")
            self.end_headers()
            self.wfile.write(mk_chunks(response_xml_string, chunk_size=self.server.chunk_size))
        else:
            self.send_header("Content-length", str(len(response_xml_string)))
            self.end_headers()
            self.wfile.write(response_xml_string)

    def do_GET(self):  # pylint: disable=invalid-name
        if self.server.dispatcher is None:
            # close this connection
            self.close_connection = True  # pylint: disable=attribute-defined-outside-init
            response_xml_string = 'received a POST request, but have no dispatcher'
            self.send_response(404, response_xml_string)  # not found
            return

        component = self.server.dispatcher.get_instance(self.get_first_path_element())

        peer_name = self.connection.getpeername()
        result = component.do_get(self.headers, self.path, peer_name)
        http_status, http_reason, response_xml_string, content_type = result

        self.send_response(http_status, http_reason)
        response_xml_string = self._compress_if_supported(response_xml_string)
        self.send_header("Content-type", content_type)
        if self.server.chunk_size > 0:
            self.send_header("transfer-encoding", "chunked")
            self.end_headers()
            self.wfile.write(mk_chunks(response_xml_string, chunk_size=self.server.chunk_size))
        else:
            self.send_header("Content-length", str(len(response_xml_string)))
            self.end_headers()
            self.wfile.write(response_xml_string)



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING

from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum

from .porttypebase import (
    ServiceWithOperations,
    WSDLMessageDescription,
    WSDLOperationBinding,
    mk_wsdl_one_way_operation,
    mk_wsdl_two_way_operation,
    msg_prefix,
)
from .stateeventserviceimpl import fill_episodic_report_body, fill_periodic_report_body

if TYPE_CHECKING:
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.mdib.statecontainers import AbstractStateContainer
    from sdc11073.provider.periodicreports import PeriodicStates


class ContextService(ServiceWithOperations):
    port_type_name = PrefixesEnum.SDC.tag('ContextService')
    WSDLMessageDescriptions = (WSDLMessageDescription('SetContextState',
                                                      (f'{msg_prefix}:SetContextState',)),
                               WSDLMessageDescription('SetContextStateResponse',
                                                      (f'{msg_prefix}:SetContextStateResponse',)),
                               WSDLMessageDescription('GetContextStates',
                                                      (f'{msg_prefix}:GetContextStates',)),
                               WSDLMessageDescription('GetContextStatesResponse',
                                                      (f'{msg_prefix}:GetContextStatesResponse',)),
                               WSDLMessageDescription('EpisodicContextReport',
                                                      (f'{msg_prefix}:EpisodicContextReport',)),
                               WSDLMessageDescription('PeriodicContextReport',
                                                      (f'{msg_prefix}:PeriodicContextReport',)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('SetContextState', 'literal', 'literal'),  # ToDo: generate wsdl:fault
                             WSDLOperationBinding('GetContextStates', 'literal', 'literal'),
                             WSDLOperationBinding('EpisodicContextReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicContextReport', None, 'literal'),
                             )

    def register_hosting_service(self, hosting_service):
        super().register_hosting_service(hosting_service)
        actions = self._mdib.sdc_definitions.Actions
        msg_names = self._mdib.sdc_definitions.data_model.msg_names
        hosting_service.register_post_handler(DispatchKey(actions.SetContextState, msg_names.SetContextState),
                                              self._on_set_context_state)
        hosting_service.register_post_handler(DispatchKey(actions.GetContextStates, msg_names.GetContextStates),
                                              self._on_get_context_states)

    def _on_set_context_state(self, request_data):
        data_model = self._sdc_definitions.data_model
        msg_node = request_data.message_data.p_msg.msg_node
        set_context_state = data_model.msg_types.SetContextState.from_node(msg_node)
        response = data_model.msg_types.SetContextStateResponse()
        return self._handle_operation_request(request_data, set_context_state, response)

    def _on_get_context_states(self, request_data):
        data_model = self._sdc_definitions.data_model
        pm_names = data_model.pm_names
        self._logger.debug('_on_get_context_states')
        msg_node = request_data.message_data.p_msg.msg_node
        get_context_state = data_model.msg_types.GetContextStates.from_node(msg_node)
        requested_handles = get_context_state.HandleRef
        if len(requested_handles) > 0:
            self._logger.info('_on_get_context_states requested Handles:{}', requested_handles)
        with self._mdib.mdib_lock:
            if len(requested_handles) == 0:
                # MessageModel: If the HANDLE reference list is empty, all states in the MDIB SHALL be included in the result list.
                context_state_containers = list(self._mdib.context_states.objects)
            else:
                context_state_containers_lookup = OrderedDict()  # lookup to avoid double entries
                for handle in requested_handles:
                    # If a HANDLE reference does match a multi state HANDLE,
                    # the corresponding multi state SHALL be included in the result list
                    tmp = self._mdib.context_states.handle.get_one(handle, allow_none=True)
                    if tmp:
                        tmp = [tmp]
                    if not tmp:
                        # If a HANDLE reference does match a descriptor HANDLE,
                        # all states that belong to the corresponding descriptor SHALL be included in the result list
                        tmp = self._mdib.context_states.descriptor_handle.get(handle)
                    if not tmp:
                        # R5042: If a HANDLE reference from the msg:GetContextStates/msg:HandleRef list does match an
                        # MDS descriptor, then all context states that are part of this MDS SHALL be included in the result list.
                        descr = self._mdib.descriptions.handle.get_one(handle, allow_none=True)
                        if descr:
                            if pm_names.MdsDescriptor == descr.NODETYPE:
                                tmp = list(self._mdib.context_states.objects)
                    if tmp:
                        for state in tmp:
                            context_state_containers_lookup[state.Handle] = state
                context_state_containers = context_state_containers_lookup.values()

        response = data_model.msg_types.GetContextStatesResponse()
        response.ContextState.extend(context_state_containers)
        response.set_mdib_version_group(self._mdib.mdib_version_group)
        response_envelope = self._sdc_device.msg_factory.mk_reply_soap_message(request_data, response)
        return response_envelope

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node, True)
        mk_wsdl_two_way_operation(port_type, operation_name='SetContextState')
        mk_wsdl_two_way_operation(port_type, operation_name='GetContextStates')
        mk_wsdl_one_way_operation(port_type, operation_name='EpisodicContextReport')
        mk_wsdl_one_way_operation(port_type, operation_name='PeriodicContextReport')

    def send_episodic_context_report(self, states: list[AbstractStateContainer],
                                     mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        nsh = data_model.ns_helper
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.EpisodicContextReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_episodic_report_body(report, states)
        ns_map = nsh.partial_map(nsh.PM, nsh.MSG, nsh.XSI, nsh.EXT, nsh.XML)
        body_node = report.as_etree_node(report.NODETYPE, ns_map)

        self._logger.debug('sending episodic context report {}', states)
        subscription_mgr.send_to_subscribers(body_node, report.action.value, mdib_version_group)

    def send_periodic_context_report(self, periodic_states_list: list[PeriodicStates],
                                     mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.PeriodicContextReport()
        report.set_mdib_version_group(mdib_version_group)
        actual_mdib_version = periodic_states_list[-1].mdib_version
        report.MdibVersion = actual_mdib_version
        fill_periodic_report_body(report, periodic_states_list)
        self._logger.debug('sending periodic context report, contains last {} episodic updates',
                           len(periodic_states_list))
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)



from __future__ import annotations

from io import BytesIO
from typing import Optional, Union, List, Type, TYPE_CHECKING

from lxml import etree

from .msgreader import validate_node
from .soapenvelope import Soap12Envelope
from sdc11073.schema_resolver import mk_schema_validator

if TYPE_CHECKING:
    from sdc11073.xml_types.addressing_types import HeaderInformationBlock
    from sdc11073.xml_types.msg_types import MessageType
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.namespaces import PrefixNamespace
    from sdc11073 import xml_utils


class CreatedMessage:
    def __init__(self, message: Soap12Envelope, msg_factory):
        self.p_msg = message
        self.msg_factory = msg_factory

    def serialize(self, pretty=False, request_manipulator=None, validate=True):
        return self.msg_factory.serialize_message(self, pretty, request_manipulator, validate)


# pylint: disable=no-self-use


class MessageFactory:
    """This class creates soap messages. It is used in two phases:
     1) call one of the mk_xxx methods. All return a CreatedMessage instance that contains the data provided in the call
     2) call the serialize method of the CreatedMessage instance to get the xml representation
     """

    def __init__(self, sdc_definitions: Type[BaseDefinitions],
                 additional_schema_specs: Union[List[PrefixNamespace], None],
                 logger,
                 validate=True):
        self.schema_specs = [entry.value for entry in sdc_definitions.data_model.ns_helper.prefix_enum]
        if additional_schema_specs is not None:
            self.schema_specs.extend(additional_schema_specs)
        self._logger = logger
        self.ns_hlp = sdc_definitions.data_model.ns_helper
        self._validate = validate
        self._xml_schema: etree.XMLSchema = mk_schema_validator(self.schema_specs, self.ns_hlp)

    def serialize_message(self, message: CreatedMessage, pretty=False,
                          request_manipulator=None, validate=True) -> bytes:
        """

        :param message: a CreatedMessage instance
        :param pretty:
        :param request_manipulator: can modify data before sending
        :param validate: if False, no validation is performed, independent of constructor setting
        :return: bytes
        """
        p_msg = message.p_msg
        nsh = self.ns_hlp
        tmp = BytesIO()
        root = etree.Element(nsh.S12.tag('Envelope'), nsmap=p_msg.nsmap)

        header_node = etree.SubElement(root, nsh.S12.tag('Header'))
        if p_msg.header_info_block:
            info_node = p_msg.header_info_block.as_etree_node('tmp', {})
            header_node.extend(info_node[:])
        header_node.extend(p_msg.header_nodes)
        body_node = etree.SubElement(root, nsh.S12.tag('Body'), nsmap=p_msg.nsmap)
        if validate:
            self._validate_node(root)
        if p_msg.payload_element is not None:
            if validate:
                self._validate_node(p_msg.payload_element)
            body_node.append(p_msg.payload_element)

        doc = etree.ElementTree(element=root)
        if hasattr(request_manipulator, 'manipulate_domtree'):
            _doc = request_manipulator.manipulate_domtree(doc)
            if _doc:
                doc = _doc
        doc.write(tmp, encoding='UTF-8', xml_declaration=True, pretty_print=pretty)
        return tmp.getvalue()

    def mk_soap_message(self,
                        header_info: HeaderInformationBlock,
                        payload: MessageType,
                        ns_list: Optional[list] = None,
                        use_defaults=True) -> CreatedMessage:
        nsh = self.ns_hlp
        if use_defaults:
            ns_set = {nsh.S12, nsh.WSA, nsh.MSG, nsh.PM}  # default
        else:
            ns_set = set()
        ns_set.update(payload.additional_namespaces)
        if ns_list:
            ns_set.update(ns_list)
        my_ns_map = nsh.partial_map(*ns_set)
        soap_envelope = Soap12Envelope(my_ns_map)
        soap_envelope.set_header_info_block(header_info)
        soap_envelope.payload_element = payload.as_etree_node(payload.NODETYPE, my_ns_map)
        return CreatedMessage(soap_envelope, self)

    def mk_soap_message_etree_payload(self,
                                      header_info: HeaderInformationBlock,
                                      payload_element: xml_utils.LxmlElement | None = None) -> CreatedMessage:
        nsh = self.ns_hlp
        my_ns_map = nsh.partial_map(nsh.S12, nsh.WSE, nsh.WSA)
        soap_envelope = Soap12Envelope(my_ns_map)
        soap_envelope.set_header_info_block(header_info)
        soap_envelope.payload_element = payload_element
        return CreatedMessage(soap_envelope, self)

    def mk_reply_soap_message(self,
                              request,
                              response_payload: MessageType,
                              ns_map: Optional[list] = None) -> CreatedMessage:
        nsh = self.ns_hlp
        ns_set = {nsh.S12, nsh.WSA, nsh.MSG, nsh.PM}  # default
        ns_set.update(response_payload.additional_namespaces)
        if ns_map:
            ns_set.update(ns_map)
        my_ns_map = nsh.partial_map(*ns_set)
        soap_envelope = Soap12Envelope(my_ns_map)
        reply_address = request.message_data.p_msg.header_info_block.mk_reply_header_block(
            action=response_payload.action)
        soap_envelope.set_header_info_block(reply_address)
        soap_envelope.payload_element = response_payload.as_etree_node(response_payload.NODETYPE, my_ns_map)
        return CreatedMessage(soap_envelope, self)

    def _validate_node(self, node):
        if self._validate:
            validate_node(node, self._xml_schema, self._logger)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from sdc11073.namespaces import PrefixesEnum

from .porttypebase import (
    DPWSPortTypeBase,
    WSDLMessageDescription,
    WSDLOperationBinding,
    mk_wsdl_one_way_operation,
    msg_prefix,
)

if TYPE_CHECKING:
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.mdib.statecontainers import AbstractStateContainer
    from sdc11073.provider.periodicreports import PeriodicStates
    from sdc11073.xml_types.msg_types import SystemErrorReportPart


class StateEventService(DPWSPortTypeBase):
    port_type_name = PrefixesEnum.SDC.tag('StateEventService')
    WSDLMessageDescriptions = (
        WSDLMessageDescription('EpisodicAlertReport',
                               (f'{msg_prefix}:EpisodicAlertReport',)),
        WSDLMessageDescription('SystemErrorReport',
                               (f'{msg_prefix}:SystemErrorReport',)),
        WSDLMessageDescription('PeriodicAlertReport',
                               (f'{msg_prefix}:PeriodicAlertReport',)),
        WSDLMessageDescription('EpisodicComponentReport',
                               (f'{msg_prefix}:EpisodicComponentReport',)),
        WSDLMessageDescription('PeriodicOperationalStateReport',
                               (f'{msg_prefix}:PeriodicOperationalStateReport',)),
        WSDLMessageDescription('PeriodicComponentReport',
                               (f'{msg_prefix}:PeriodicComponentReport',)),
        WSDLMessageDescription('EpisodicOperationalStateReport',
                               (f'{msg_prefix}:EpisodicOperationalStateReport',)),
        WSDLMessageDescription('PeriodicMetricReport',
                               (f'{msg_prefix}:PeriodicMetricReport',)),
        WSDLMessageDescription('EpisodicMetricReport',
                               (f'{msg_prefix}:EpisodicMetricReport',)),
    )

    WSDLOperationBindings = (WSDLOperationBinding('EpisodicAlertReport', None, 'literal'),
                             WSDLOperationBinding('SystemErrorReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicAlertReport', None, 'literal'),
                             WSDLOperationBinding('EpisodicComponentReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicOperationalStateReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicComponentReport', None, 'literal'),
                             WSDLOperationBinding('EpisodicOperationalStateReport', None, 'literal'),
                             WSDLOperationBinding('PeriodicMetricReport', None, 'literal'),
                             WSDLOperationBinding('EpisodicMetricReport', None, 'literal'),
                             )

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node, True)
        mk_wsdl_one_way_operation(port_type, operation_name='EpisodicAlertReport')
        mk_wsdl_one_way_operation(port_type, operation_name='SystemErrorReport')
        mk_wsdl_one_way_operation(port_type, operation_name='PeriodicAlertReport')
        mk_wsdl_one_way_operation(port_type, operation_name='EpisodicComponentReport')
        mk_wsdl_one_way_operation(port_type, operation_name='PeriodicOperationalStateReport')
        mk_wsdl_one_way_operation(port_type, operation_name='PeriodicComponentReport')
        mk_wsdl_one_way_operation(port_type, operation_name='EpisodicOperationalStateReport')
        mk_wsdl_one_way_operation(port_type, operation_name='PeriodicMetricReport')
        mk_wsdl_one_way_operation(port_type, operation_name='EpisodicMetricReport')

    def send_episodic_metric_report(self, states: list[AbstractStateContainer],
                                    mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.EpisodicMetricReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_episodic_report_body(report, states)
        self._logger.debug('sending episodic metric report {}', states)
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_periodic_metric_report(self, periodic_states_list: list[PeriodicStates],
                                    mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.PeriodicMetricReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_periodic_report_body(report, periodic_states_list)
        self._logger.debug('sending periodic metric report, contains last {} episodic updates',
                           len(periodic_states_list))
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_episodic_alert_report(self, states: list[AbstractStateContainer],
                                   mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.EpisodicAlertReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_episodic_report_body(report, states)
        self._logger.debug('sending episodic alert report {}', states)
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_periodic_alert_report(self, periodic_states_list: list[PeriodicStates],
                                   mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.PeriodicAlertReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_periodic_report_body(report, periodic_states_list)
        self._logger.debug('sending periodic alert report, contains last {} episodic updates',
                           len(periodic_states_list))
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_episodic_operational_state_report(self, states: list[AbstractStateContainer],
                                               mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.EpisodicOperationalStateReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_episodic_report_body(report, states)
        self._logger.debug('sending episodic operational state report {}', states)
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_periodic_operational_state_report(self, periodic_states_list: list[PeriodicStates],
                                               mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.PeriodicOperationalStateReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_periodic_report_body(report, periodic_states_list)
        self._logger.debug('sending periodic operational state report, contains last {} episodic updates',
                           len(periodic_states_list))
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_episodic_component_state_report(self, states: list[AbstractStateContainer],
                                             mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.EpisodicComponentReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_episodic_report_body(report, states)
        self._logger.debug('sending episodic component report {}', states)
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_periodic_component_state_report(self, periodic_states_list: list[PeriodicStates],
                                             mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.PeriodicComponentReport()
        report.set_mdib_version_group(mdib_version_group)
        fill_periodic_report_body(report, periodic_states_list)
        self._logger.debug('sending periodic component report, contains last {} episodic updates',
                           len(periodic_states_list))
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

    def send_system_error_report(self, report_parts: list[SystemErrorReportPart],
                                 mdib_version_group: MdibVersionGroup):
        data_model = self._sdc_definitions.data_model
        subscription_mgr = self.hosting_service.subscriptions_manager
        report = data_model.msg_types.SystemErrorReport()
        report.ReportPart.extend(report_parts)
        report.set_mdib_version_group(mdib_version_group)
        self._logger.debug('sending SystemErrorReport')
        subscription_mgr.send_to_subscribers(report, report.action.value, mdib_version_group)

def fill_episodic_report_body(report, states):
    """Helper that splits states list into separate lists per source mds and adds them to report accordingly."""
    lookup = _separate_states_by_source_mds(states)
    for source_mds_handle, states in lookup.items():
        report_part = report.add_report_part()
        report_part.SourceMds = source_mds_handle
        report_part.values_list.extend(states)


def fill_periodic_report_body(report, report_parts):
    for tmp in report_parts:
        lookup = _separate_states_by_source_mds(tmp.states)
        for source_mds_handle, states in lookup.items():
            report_part = report.add_report_part()
            report_part.SourceMds = source_mds_handle
            report_part.values_list.extend(states)


def _separate_states_by_source_mds(states) -> dict:
    lookup = defaultdict(list)
    for state in states:
        lookup[state.source_mds].append(state)
    if None in lookup:
        raise ValueError(f'States {[st.DescriptorHandle for st in lookup[None]]} have no source mds')
    return lookup



"""Implementation of patient provider functionality."""
from __future__ import annotations

from typing import TYPE_CHECKING

from .contextprovider import GenericContextProvider

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.provider.operations import OperationDefinitionBase
    from sdc11073.provider.sco import AbstractScoOperationsRegistry

    from .providerbase import OperationClassGetter


class GenericPatientContextProvider(GenericContextProvider):
    """Example for handling of SetContextState operations.

    This Provider instantiates a SetContextState operation if the operation target is a PatientContextDescriptor.
    Nothing is added to the mdib. If the mdib does not contain these operations, the functionality is not available.
    """

    def __init__(self, mdib: ProviderMdibProtocol, log_prefix: str | None):
        super().__init__(mdib, log_prefix=log_prefix)
        self._patient_context_entity = None
        self._set_patient_context_operations = []

    def init_operations(self, sco: AbstractScoOperationsRegistry):
        """Find the PatientContextDescriptor."""
        super().init_operations(sco)
        pm_names = self._mdib.data_model.pm_names
        entities = self._mdib.entities.by_node_type(pm_names.PatientContextDescriptor)
        if len(entities) == 1:
            self._patient_context_entity = entities[0]

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Add Operation Handler if operation target is the previously found PatientContextDescriptor."""
        if self._patient_context_entity is not None and \
                operation_descriptor_container.OperationTarget == self._patient_context_entity.handle:
            pc_operation = self._mk_operation_from_operation_descriptor(operation_descriptor_container,
                                                                        operation_cls_getter,
                                                                        operation_handler=self._set_context_state)
            self._set_patient_context_operations.append(pc_operation)
            return pc_operation
        return None


class PatientContextProvider(GenericPatientContextProvider):
    """PatientContextProvider adds operations to mdib if they do not exist."""

    def make_missing_operations(self, sco: AbstractScoOperationsRegistry) -> list[OperationDefinitionBase]:
        """Add operation to mdib if it does not exist."""
        pm_names = self._mdib.data_model.pm_names
        ops = []
        operation_cls_getter = sco.operation_cls_getter
        if self._patient_context_entity is not None  and not self._set_patient_context_operations:
            op_cls = operation_cls_getter(pm_names.SetContextStateOperationDescriptor)
            pc_operation = op_cls('opSetPatCtx',
                                  self._patient_context_entity.handle,
                                  self._set_context_state,
                                  coded_value=None)
            ops.append(pc_operation)
        return ops



from __future__ import annotations

import typing
from io import BytesIO

from lxml import etree

from sdc11073.dispatch import DispatchKey, RequestDispatcher
from sdc11073.namespaces import EventingActions
from sdc11073.namespaces import default_ns_helper as ns_hlp
from sdc11073.xml_types import mex_types
from sdc11073.xml_types.addressing_types import EndpointReferenceType
from sdc11073.xml_types.dpws_types import HostedServiceType

if typing.TYPE_CHECKING:
    import pathlib
    from sdc11073 import xml_utils

_wsdl_ns = ns_hlp.WSDL.namespace

WSP_NS = ns_hlp.WSP.namespace
_WSP_PREFIX = ns_hlp.WSP.prefix

# DiscoveryType, only used in SDC
_DISCOVERY_TYPE_NS = "http://standards.ieee.org/downloads/11073/11073-10207-2017"

WSDL_S12 = ns_hlp.WSDL12.namespace  # old soap 12 namespace, used in wsdl 1.1. used only for wsdl


def etree_from_file(path: str | pathlib.Path) -> xml_utils.LxmlElement:
    parser = etree.ETCompatXMLParser(resolve_entities=False)
    doc = etree.parse(str(path), parser=parser)
    return doc.getroot()


class _EventService(RequestDispatcher):
    """A service that offers subscriptions."""

    def __init__(self, sdc_device, subscriptions_manager, offered_subscriptions):
        super().__init__()
        self._msg_reader = sdc_device.msg_reader
        self._subscriptions_manager = subscriptions_manager
        self._offered_subscriptions = offered_subscriptions
        self.register_post_handler(DispatchKey(EventingActions.Subscribe, ns_hlp.WSE.tag('Subscribe')),
                                   self._on_subscribe)
        self.register_post_handler(DispatchKey(EventingActions.Unsubscribe, ns_hlp.WSE.tag('Unsubscribe')),
                                   self._on_unsubscribe)
        self.register_post_handler(DispatchKey(EventingActions.GetStatus, ns_hlp.WSE.tag('GetStatus')),
                                   self._on_get_status)
        self.register_post_handler(DispatchKey(EventingActions.Renew, ns_hlp.WSE.tag('Renew')),
                                   self._on_renew_status)

    @property
    def subscriptions_manager(self):
        return self._subscriptions_manager

    def _on_subscribe(self, request_data):
        returned_envelope = self._subscriptions_manager.on_subscribe_request(request_data)
        return returned_envelope

    def _on_unsubscribe(self, request_data):
        returned_envelope = self._subscriptions_manager.on_unsubscribe_request(request_data)
        return returned_envelope

    def _on_get_status(self, request_data):
        returned_envelope = self._subscriptions_manager.on_get_status_request(request_data)
        return returned_envelope

    def _on_renew_status(self, request_data):
        returned_envelope = self._subscriptions_manager.on_renew_request(request_data)
        return returned_envelope


class DPWSHostedService(_EventService):
    """Container for DPWSPortTypeBase instances."""

    def __init__(self, sdc_device, subscriptions_manager, path_element, port_type_impls):
        """:param sdc_device:
        :param path_element:
        :param port_type_impls: list of DPWSPortTypeBase
        """
        offered_subscriptions = []
        for p in port_type_impls:
            offered_subscriptions.extend(p.offered_subscriptions)
        super().__init__(sdc_device, subscriptions_manager, offered_subscriptions)
        self.path_element = path_element

        self._sdc_device = sdc_device
        self._mdib = sdc_device.mdib
        self.port_type_impls = port_type_impls
        self._wsdl_string = self._mk_wsdl_string()
        self.register_post_handler(DispatchKey(f'{ns_hlp.WSX.namespace}/GetMetadata/Request',
                                               ns_hlp.WSX.tag('GetMetadata')),
                                   self._on_get_metadata)
        self.register_get_handler('?wsdl', self._on_get_wsdl)
        for port_type_impl in port_type_impls:
            port_type_impl.register_hosting_service(self)

    def mk_dpws_hosted_instance(self) -> HostedServiceType:
        endpoint_references_list = []
        for addr in self._sdc_device.base_urls:
            epr_type = EndpointReferenceType()
            epr_type.Address = f'{addr.geturl()}/{self.path_element}'
            endpoint_references_list.append(epr_type)
        dpws_hosted = HostedServiceType()
        dpws_hosted.EndpointReference.extend(endpoint_references_list)
        dpws_hosted.Types = [p.port_type_name for p in self.port_type_impls]
        dpws_hosted.ServiceId = self.path_element  # value seems to be not important as long as it is unique
        return dpws_hosted

    def _on_get_wsdl(self) -> bytes:
        """Return wsdl."""
        self._logger.debug('_onGetWsdl returns {}', self._wsdl_string)
        return self._wsdl_string

    def _mk_wsdl_string(self) -> bytes:
        sdc_definitions = self._sdc_device.mdib.sdc_definitions
        my_nsmap = ns_hlp.partial_map(
            ns_hlp.MSG, ns_hlp.PM, ns_hlp.WSA, ns_hlp.WSE, ns_hlp.DPWS, ns_hlp.MDPWS)
        my_nsmap['tns'] = ns_hlp.SDC.namespace
        my_nsmap['dt'] = _DISCOVERY_TYPE_NS
        porttype_prefix = 'tns'
        my_nsmap['wsdl'] = _wsdl_ns
        my_nsmap['s12'] = WSDL_S12
        my_nsmap[_WSP_PREFIX] = WSP_NS
        for port_type_impl in self.port_type_impls:
            for entry in port_type_impl.additional_namespaces:
                my_nsmap[entry.prefix] = entry.namespace
        wsdl_definitions = etree.Element(etree.QName(_wsdl_ns, 'definitions'),
                                         nsmap=my_nsmap,
                                         attrib={'targetNamespace': sdc_definitions.PortTypeNamespace})

        types = etree.SubElement(wsdl_definitions, etree.QName(_wsdl_ns, 'types'))
        # remove annotations from schemas, this reduces wsdl size from 280kb to 100kb!
        ext_schema_ = etree_from_file(ns_hlp.EXT.local_schema_file)
        ext_schema = self._remove_annotations(ext_schema_)
        pm_schema_ = etree_from_file(ns_hlp.PM.local_schema_file)
        participant_schema = self._remove_annotations(pm_schema_)
        mm_schema_ = etree_from_file(ns_hlp.MSG.local_schema_file)
        message_schema = self._remove_annotations(mm_schema_)
        types.append(ext_schema)
        types.append(participant_schema)
        types.append(message_schema)
        # append all message nodes
        for _port_type_impl in self.port_type_impls:
            _port_type_impl.add_wsdl_messages(wsdl_definitions)
        for _port_type_impl in self.port_type_impls:
            _port_type_impl.add_wsdl_port_type(wsdl_definitions)
        for _port_type_impl in self.port_type_impls:
            _port_type_impl.add_wsdl_binding(wsdl_definitions, porttype_prefix)
        return etree.tostring(wsdl_definitions, encoding='UTF-8', xml_declaration=True)

    @staticmethod
    def _remove_annotations(root_node):
        remove_annotations_string = b"""<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                                      xmlns:xs="http://www.w3.org/2001/XMLSchema">
          <xsl:output method="xml" indent="yes"/>

          <xsl:template match="@* | node()">
            <xsl:copy>
              <xsl:apply-templates select="@* | node()"/>
            </xsl:copy>
          </xsl:template>

          <xsl:template match="xs:annotation" />
        </xsl:stylesheet>"""
        remove_annotations_doc = etree.parse(BytesIO(remove_annotations_string))
        remove_annotations_xslt = etree.XSLT(remove_annotations_doc)
        return remove_annotations_xslt(root_node).getroot()

    def _on_get_metadata(self, request_data):
        msg_factory = self._sdc_device.msg_factory
        consumed_path_elements = request_data.consumed_path_elements
        http_header = request_data.http_header

        # determine the correct location of wsdl, depending on call
        host = http_header['Host']  # this is the address that was called.
        all_base_urls = self._sdc_device.base_urls
        my_base_urls = [u for u in all_base_urls if u.netloc == host]
        my_base_url = my_base_urls[0] if len(my_base_urls) > 0 else all_base_urls[0]
        tmp = '/'.join(consumed_path_elements)
        location_text = f'{my_base_url.scheme}://{my_base_url.netloc}/{tmp}/?wsdl'

        metadata = mex_types.Metadata()
        section = mex_types.RelationshipMetadataSection()
        section.MetadataReference.Host = self._sdc_device.dpws_host
        hosted = self.mk_dpws_hosted_instance()
        section.MetadataReference.Hosted.append(hosted)
        metadata.MetadataSection.append(section)

        section = mex_types.LocationMetadataSection()
        section.Location = location_text
        metadata.MetadataSection.append(section)

        # find namespaces that are used in Types of Host and Hosted
        _nsm = self._mdib.nsmapper
        needed_namespaces = [_nsm.DPWS, _nsm.WSX]
        q_names = self._sdc_device.dpws_host.Types[:]
        q_names.extend(hosted.Types)
        for q_name in q_names:
            for e in _nsm.prefix_enum:
                if e.namespace == q_name.namespace and e not in needed_namespaces:
                    needed_namespaces.append(e)
        response = msg_factory.mk_reply_soap_message(request_data, metadata, needed_namespaces)
        return response

    def __repr__(self):
        return f'{self.__class__.__name__} path={self.path_element} ' \
               f'port types={[dp.port_type_string for dp in self.port_type_impls]}'



from sdc11073.roles import providerbase


class OperationProvider(providerbase.ProviderRole):
    """Handle operations that work on operation states.

     Empty implementation, not needed/used for sdc11073 tests.
     """



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lxml import etree

    from sdc11073.xml_types import wsd_types


class Service:
    """Service objects contain discovery relevant data of a service.

    They are used for publishing a service or as result (list) when searching for services in the network.
    """

    def __init__(self,
                 types: list[etree.QName] | None,
                 scopes: wsd_types.ScopesType | None,
                 x_addrs: list[str] | None,
                 epr: str,
                 instance_id: str,
                 metadata_version: int = 1):
        self.types = types
        self.scopes = scopes
        self._x_addrs = x_addrs
        self.epr = epr
        self.instance_id = instance_id
        self.message_number = 0
        self.metadata_version = metadata_version

    @property
    def x_addrs(self) -> list[str]:
        """Get the addresses of the service."""
        return self._x_addrs or []

    @x_addrs.setter
    def x_addrs(self, x_addrs: list[str]):
        """Set the addresses of the service."""
        self._x_addrs = x_addrs

    def increment_message_number(self) -> None:
        """Add one."""
        self.message_number += 1

    def __repr__(self) -> str:
        scopes_str = 'None' if self.scopes is None else ', '.join([str(x) for x in self.scopes.text])
        types_str = 'None' if self.types is None else ', '.join([str(x) for x in self.types])
        return f'Service epr={self.epr}, instanceId={self.instance_id} Xaddr={self._x_addrs} ' \
               f'scopes={scopes_str} types={types_str}'

    def __str__(self) -> str:
        scopes_str = 'None' if self.scopes is None else ', '.join([str(x) for x in self.scopes.text])
        types_str = 'None' if self.types is None else ', '.join([str(x) for x in self.types])
        return f'Service epr={self.epr}, instanceId={self.instance_id}\n' \
               f'   Xaddr={self._x_addrs}\n' \
               f'   scopes={scopes_str}\n' \
               f'   types={types_str}'



"""The module implements the base class for consumer and provider specific mdib implementations."""
from __future__ import annotations

from threading import RLock
from typing import TYPE_CHECKING

from sdc11073 import observableproperties as properties
from sdc11073.mdib.mdibbase import MdibVersionGroup

if TYPE_CHECKING:
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.loghelper import LoggerAdapter


class EntityMdibBase:
    """Base class for consumer and provider specific mdib implementations."""

    metric_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    waveform_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    alert_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    context_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    component_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    new_descriptors_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    updated_descriptors_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    deleted_descriptors_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    operation_handles = properties.ObservableProperty(fire_only_on_changed_value=False)
    sequence_id = properties.ObservableProperty()
    instance_id = properties.ObservableProperty()

    def __init__(self, sdc_definitions: type[BaseDefinitions],
                 logger: LoggerAdapter):
        """Construct MdibBase.

        :param sdc_definitions: a class derived from BaseDefinitions
        """
        self.sdc_definitions = sdc_definitions
        self.data_model = sdc_definitions.data_model
        self._logger = logger
        self.mdib_version = 0
        self.sequence_id = ''  # needs to be set to a reasonable value by derived class
        self.instance_id = None  # None or an unsigned int
        self.log_prefix = ''
        self.mdib_lock = RLock()

    @property
    def mdib_version_group(self) -> MdibVersionGroup:
        """"Get current version data."""
        return MdibVersionGroup(self.mdib_version, self.sequence_id, self.instance_id)

    @property
    def logger(self) -> LoggerAdapter:
        """Return the logger."""
        return self._logger



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from sdc11073 import xml_utils

    from sdc11073.pysoap.soapenvelope import Soap12Envelope


class RequestManipulatorProtocol(Protocol):
    """A request manipulator can be used to inspect or manipulate output created by the sdc client.

    Output creation is done in three steps: first a Soap12Envelope is created,
    then a (libxml) etree is created from its content, anf finally a bytestring is generated from the etree.
    The sdc client calls corresponding methods of the manipulator object after every step.
    If the method returns something different from None, this returned value will be used as input for the next step.
    """

    def manipulate_soapenvelope(self, soap_envelope: Soap12Envelope) -> Soap12Envelope | None:
        """Manipulate on Soap12Envelope level."""

    def manipulate_domtree(self, domtree: xml_utils.LxmlElement) -> xml_utils.LxmlElement | None:
        """Manipulate on etree.Element level."""

    def manipulate_string(self, xml_string: str) -> str | None:
        """Manipulate on string level."""



from __future__ import annotations

import re
from collections import namedtuple
from datetime import date, datetime, timedelta, tzinfo
from decimal import Decimal
from typing import Union

ISO8601_PERIOD_REGEX = re.compile(
    r"^(?P<sign>[+-])?"
    r"P(?!\b)"
    r"(?P<years>[0-9]+([,.][0-9]+)?Y)?"
    r"(?P<months>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<weeks>[0-9]+([,.][0-9]+)?W)?"
    r"(?P<days>[0-9]+([,.][0-9]+)?D)?"
    r"((?P<separator>T)(?P<hours>[0-9]+([,.][0-9]+)?H)?"
    r"(?P<minutes>[0-9]+([,.][0-9]+)?M)?"
    r"(?P<seconds>[0-9]+([,.][0-9]+)?S)?)?$")

# regular expression to parse ISO duration strings.

DurationType = Union[Decimal, int, float]
ParsedDurationType = Union[int, float]

def parse_duration(date_string: str) -> ParsedDurationType:
    """Parse an ISO 8601 durations into a float value containing seconds.

    The following duration formats are supported:
      -PnnW                  duration in weeks
      -PnnYnnMnnDTnnHnnMnnS  complete duration specification
    Years and month are not supported, values must be zero!
    """
    if not isinstance(date_string, str):
        raise TypeError(f"Expecting a string {date_string}")
    match = ISO8601_PERIOD_REGEX.match(date_string)
    if not match:
        raise ValueError(f"Unable to parse duration string {date_string}")
    groups = match.groupdict()
    for key, val in groups.items():
        if key not in ('separator', 'sign'):
            if val is None:
                groups[key] = "0n"
            if key in ('years', 'months'):
                groups[key] = Decimal(groups[key][:-1].replace(',', '.'))
            else:
                # these values are passed into a timedelta object,
                # which works with floats.
                groups[key] = float(groups[key][:-1].replace(',', '.'))
    if groups["years"] != 0 or groups["months"] != 0:
        raise ValueError(f"Unable to parse duration string {date_string} (Non zero year or month)")
    ret = timedelta(days=groups["days"], hours=groups["hours"],
                    minutes=groups["minutes"], seconds=groups["seconds"],
                    weeks=groups["weeks"])
    if groups["sign"] == '-':
        ret = timedelta(0) - ret
    return ret.total_seconds()


def duration_string(seconds: DurationType) -> str:
    """Create an ISO 8601 durations value containing seconds."""
    sign = '-' if seconds < 0 else ''
    fraction = abs(seconds - int(seconds))
    seconds = abs(int(seconds))
    minutes, sec = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)
    if fraction == 0:
        fraction_string = ''
    else:
        fraction_string = f'{fraction:.9f}'[1:]  # starting from dot char
        while fraction_string[-1] == '0':
            fraction_string = fraction_string[:-1]
    if days == 0:
        return f'{sign}PT{hours}H{minutes}M{sec}{fraction_string}S'
    return f'{sign}P0Y0M{days}DT{hours}H{minutes}M{sec}{fraction_string}S'


##### Date Time ######
GYearMonth = namedtuple('GYearMonth', 'year month')
GYear = namedtuple('GYear', 'year')

ZERO = timedelta(0)


class UTC(tzinfo):
    """Fixed offset in minutes east from UTC."""

    def __init__(self, offset_minutes: int, tzname: str | None = None):
        self._offset = timedelta(minutes=offset_minutes)
        self._tzname = tzname

    def utcoffset(self, dt):  # pylint:disable=unused-argument
        return self._offset

    def tzname(self, dt):  # pylint:disable=unused-argument
        return self._tzname

    def dst(self, dt):  # pylint:disable=unused-argument
        return ZERO


# regular expression to parse ISO 8601 date / datetime strings.

_TZ_REGEX_STR = '(?P<tz>Z|((?P<tz_sign>[+-])(?P<tz_hours>[0-9]{1,2}):(?P<tz_minutes>[0-9]{1,2})))?'
_DATE_REGEX_STR = '(?P<year>[0-9]{4})-?(?P<month>1[0-2]|0[1-9])-?(?P<day>[0-9]{1,2})'
_TIME_REGEX_STR = '(?P<hour>2[0-3]|[01][0-9]):?(?P<minute>[0-5][0-9]):(?P<second>[0-5][0-9]([.][0-9]+)?)'
_DATETIME_REGEX = re.compile('^' + _DATE_REGEX_STR + '(T' + _TIME_REGEX_STR + _TZ_REGEX_STR + ')?')
_DATETIME_REGEX_RELAXED = re.compile(
    '^' + _DATE_REGEX_STR + '([T, ]' + _TIME_REGEX_STR + _TZ_REGEX_STR + ')?')  # allows space between date and time
_year_month_regex = re.compile('^(?P<year>[0-9]{4})(-(?P<month>1[0-2]|0[1-9]))?')


DateTypeUnion = Union[GYear, GYearMonth, date, datetime]

def parse_date_time(date_time_str: str, strict: bool = True) -> DateTypeUnion | None:
    """Parse a date time string.

    String can be  xsd:dateTime, xsd:date, xsd:gYearMonth or xsd:gYear.
    """
    try:
        d_t = _DATETIME_REGEX.match(date_time_str) if strict else _DATETIME_REGEX_RELAXED.match(date_time_str)
        if d_t is not None:
            groups = d_t.groupdict()
            year, month, day = int(groups['year']), int(groups['month']), int(groups['day'])
            if groups['hour'] is None:  # only a date, no time
                return date(year, month, day)

            tz_1st = groups['tz']
            tz_info = None
            if tz_1st is not None:
                if tz_1st == 'Z':
                    tz_info = UTC(0, 'UTC')
                elif tz_1st[0] in ('+', '-'):
                    tz_hours = int(groups['tz_hours'])
                    tz_minutes = int(groups['tz_minutes'])
                    offset = tz_hours * 60 + tz_minutes
                    if tz_1st[0] == '-':
                        offset *= -1
                    tz_info = UTC(offset, 'unknown')

            hour = int(groups['hour'])
            minute = int(groups.get('minute', '00'))
            second = float(groups.get('second', '0.0'))
            sec, micro_sec = int(second), int((second - int(second)) * 1000000)
            value = datetime(year, month, day, hour, minute, sec, micro_sec, tz_info)
            return value

        d_t = _year_month_regex.match(date_time_str)
        if d_t is not None:
            groups = d_t.groupdict()
            year, month = groups['year'], groups['month']
            value = GYear(int(year)) if month is None else GYearMonth(int(year), int(month))
            return value

        raise ValueError(f'Could not parse date string = "{date_time_string}"')
    except ValueError:
        return None


def _mk_seconds_string(date_object: DateTypeUnion) -> str:
    if date_object.microsecond > 0:
        seconds = float(date_object.second) + float(date_object.microsecond) / 1e6
        seconds_string = f'{seconds:06.03f}'
        # remove trailing zeros
        while seconds_string[-1] == '0':
            seconds_string = seconds_string[:-1]
    else:
        seconds_string = f'{date_object.second:02d}'
    return seconds_string


def _mk_tz_string(date_object: DateTypeUnion) -> str:
    tz_string = ''
    if date_object.tzinfo:
        delta = date_object.tzinfo.utcoffset(0)
        tz_seconds = delta.seconds + (3600 * 24) * delta.days
        if tz_seconds == 0:
            tz_string = 'Z'
        if tz_seconds != 0:
            minutes, _ = divmod(abs(tz_seconds), 60)
            hours, minutes = divmod(minutes, 60)
            sign = '+' if tz_seconds > 0 else '-'
            tz_string = f'{sign}{hours:02d}:{minutes:02d}'
    return tz_string


def date_time_string(date_object: DateTypeUnion) -> str:
    if hasattr(date_object, 'hour'):  # datetime object
        date_string = '{:4d}-{:02d}-{:02d}T{:02d}:{:02d}:{}{}'.format(
            date_object.year, date_object.month, date_object.day,
            date_object.hour, date_object.minute,
            _mk_seconds_string(date_object), _mk_tz_string(date_object))
    elif hasattr(date_object, 'day'):  # date object
        date_string = f'{date_object.year:4d}-{date_object.month:02d}-{date_object.day:02d}'
    elif hasattr(date_object, 'month'):  # GYearMonth object
        date_string = f'{date_object.year:4d}-{date_object.month:02d}'
    elif hasattr(date_object, 'year'):  # GYear object
        date_string = f'{date_object.year:4d}'
    else:
        raise ValueError(f'cannot convert {date_object.__class__.__name__} to ISO8601 datetime string')
    return date_string



"""Common definitions for WS-Discovery module."""

import logging

from sdc11073.definitions_sdc import SdcV1Definitions
from sdc11073.pysoap.msgfactory import MessageFactory
from sdc11073.pysoap.msgreader import MessageReader

message_factory = MessageFactory(SdcV1Definitions, None, logger=logging.getLogger('sdc.discover.msg'))
message_reader = MessageReader(SdcV1Definitions, None, logger=logging.getLogger('sdc.discover.msg'))

MULTICAST_PORT = 3702
MULTICAST_IPV4_ADDRESS = "239.255.255.250"



from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING

from aiohttp.client import ClientSession, ClientTimeout, TCPConnector

from sdc11073 import commlog, observableproperties
from sdc11073.httpserver.compression import CompressionHandler
from sdc11073.httpserver.httpreader import mk_chunks
from sdc11073.namespaces import default_ns_helper as ns_hlp
from sdc11073.pysoap.soapenvelope import Fault

from .soapclient import HTTPReturnCodeError

if TYPE_CHECKING:
    from ssl import SSLContext

    from sdc11073.consumer.manipulator import RequestManipulatorProtocol
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.pysoap.msgreader import MessageReader, ReceivedMessage


class SoapClientAsync:
    """SOAP Client wraps an http connection. It can send / receive SoapEnvelopes."""

    _used_soap_clients = 0

    roundtrip_time = observableproperties.ObservableProperty()

    def __init__(self,
                 netloc: str,
                 socket_timeout: int | float,
                 logger: LoggerAdapter,
                 ssl_context: [SSLContext, None],
                 sdc_definitions: BaseDefinitions,
                 msg_reader: MessageReader,
                 supported_encodings: list[str] | None = None,
                 request_encodings: list[str] | None = None,
                 chunk_size: int = 0):
        self._log = logger
        self._ssl_context = ssl_context
        self._sdc_definitions = sdc_definitions
        self._msg_reader = msg_reader
        self._netloc = netloc
        self._socket_timeout = socket_timeout
        self._http_connection = None  # connect later on demand
        self.__class__._used_soap_clients += 1  # noqa: SLF001
        self._client_number = self.__class__._used_soap_clients  # noqa: SLF001
        self._log.info('created soap client No. {} for {}', self._client_number, netloc)
        self.supported_encodings = supported_encodings if supported_encodings is not None \
            else CompressionHandler.available_encodings
        # these compression alg's does the other side accept ( set at runtime):
        self.request_encodings = request_encodings if request_encodings is not None else []
        self._get_headers = self._make_get_headers()
        self._chunk_size = chunk_size
        self._netloc = netloc

    @property
    def netloc(self) -> str:
        """Return location, e.g.127.0.0.1:9999."""
        return self._netloc

    def is_closed(self) -> bool:
        """Return True if connection is closed."""
        return self._http_connection is None

    async def _mk_http_connection(self) -> ClientSession:
        """TCP_NODELAY is set by default in asyncio create_connection."""
        if self._ssl_context is not None:
            connector = TCPConnector(ssl=self._ssl_context)
            base_url = f'https://{self._netloc}/'
        else:
            connector = TCPConnector()
            base_url = f'http://{self._netloc}/'

        return ClientSession(base_url, connector=connector, timeout=ClientTimeout(self._socket_timeout))

    async def async_connect(self):
        """Connect to netloc."""
        self._http_connection = await self._mk_http_connection()

    def close(self):
        """Close connection."""
        # ToDo: run async_close in event loop
        self._http_connection = None

    async def async_close(self):
        """Close connection."""
        if self._http_connection is not None:
            self._log.info('closing soapClientNo {} for {}', self._client_number, self._netloc)
            await self._http_connection.close()
            self._http_connection = None

    async def async_post_message_to(self, path: str,
                                    created_message: CreatedMessage,
                                    request_manipulator: RequestManipulatorProtocol | None = None) \
            -> ReceivedMessage | None:
        """Send the message and return None if the response is empty else the received response.

        :param path: url path component
        :param created_message: The message that shall be sent
        :param request_manipulator: can manipulate data before sending
        """
        if self.is_closed():
            self._http_connection = await self._mk_http_connection()

        if hasattr(request_manipulator, 'manipulate_soapenvelope'):
            tmp = request_manipulator.manipulate_soapenvelope(created_message.p_msg)
            if tmp:
                created_message.p_msg = tmp
        xml_request = created_message.serialize(request_manipulator=request_manipulator)

        assert b'utf-8' in xml_request[:100].lower()
        if hasattr(request_manipulator, 'manipulate_string'):
            tmp = request_manipulator.manipulate_string(xml_request)
            if tmp:
                xml_request = tmp

        started = time.perf_counter()
        try:
            headers = {
                'Content-type': 'application/soap+xml; charset=utf-8',
                'user_agent': 'pysoap',
                'Connection': 'keep-alive',
            }
            logging.getLogger(commlog.SOAP_REQUEST_OUT).debug(xml_request, extra={'http_method': 'POST'})

            if self.supported_encodings:
                headers['Accept-Encoding'] = ','.join(self.supported_encodings)
            if self.request_encodings:
                for compr in self.request_encodings:
                    if compr in self.supported_encodings:
                        xml_request = CompressionHandler.compress_payload(compr, xml_request)
                        headers['Content-Encoding'] = compr
                        break
            if self._chunk_size > 0:
                headers['transfer-encoding'] = "chunked"
                xml_request = mk_chunks(xml_request, chunk_size=self._chunk_size)
            else:
                headers['Content-Length'] = str(len(xml_request))

            async with self._http_connection.post(path, data=xml_request, headers=headers) as resp:
                xml_response = await resp.text()

        finally:
            self.roundtrip_time = time.perf_counter() - started  # set roundtrip time even if method raises an exception
        if not xml_response:  # empty response
            return None

        message_data = self._msg_reader.read_received_message(xml_response.encode('utf-8'))
        if message_data.action == f'{ns_hlp.WSA.namespace}/fault':
            soap_fault = Fault.from_node(message_data.p_msg.msg_node)
            raise HTTPReturnCodeError(resp.status, resp.reason, soap_fault)
        return message_data

    def _make_get_headers(self) -> dict[str, str]:
        headers = {
            'user_agent': 'pysoap',
            'Connection': 'keep-alive',
        }
        if self.supported_encodings:
            headers['Accept-Encoding'] = ', '.join(self.supported_encodings)
        return headers



"""The module contains extensions to the functionality of the EntityProviderMdib."""
from __future__ import annotations

import time
from collections import defaultdict
from typing import TYPE_CHECKING

from sdc11073.etc import apply_map
from sdc11073.exceptions import ApiUsageError
from sdc11073.xml_types.pm_types import RetrievabilityMethod

if TYPE_CHECKING:
    from sdc11073.location import SdcLocation
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorContainer
    from sdc11073.mdib.statecontainers import AbstractStateContainer
    from sdc11073.xml_types.pm_types import InstanceIdentifier

    from .entities import ProviderEntity, ProviderMultiStateEntity
    from .entity_providermdib import EntityProviderMdib


class EntityProviderMdibMethods:
    """Extra methods for provider mdib that are not part of core functionality."""

    def __init__(self, provider_mdib: EntityProviderMdib):
        self._mdib = provider_mdib
        self.default_validators = (provider_mdib.data_model.pm_types.InstanceIdentifier(
            root='rootWithNoMeaning', extension_string='System'),)

    def set_all_source_mds(self):
        """Set source mds in all entities."""
        dict_by_parent_handle = defaultdict(list)
        descriptor_containers = [entity.descriptor for entity in self._mdib.internal_entities.values()]
        for d in descriptor_containers:
            dict_by_parent_handle[d.parent_handle].append(d)

        def tag_tree(source_mds_handle: str, descriptor_container: AbstractDescriptorContainer):
            descriptor_container.set_source_mds(source_mds_handle)
            children = dict_by_parent_handle[descriptor_container.Handle]
            for ch in children:
                tag_tree(source_mds_handle, ch)

        for mds in dict_by_parent_handle[None]:  # only mds has no parent
            tag_tree(mds.Handle, mds)

    def set_location(self, sdc_location: SdcLocation,
                     validators: list[InstanceIdentifier] | None = None,
                     location_context_descriptor_handle: str | None = None):
        """Create a location context state. The new state will be the associated state.

        This method updates only the mdib data!
        Use the SdcProvider.set_location method if you want to publish the address on the network.
        :param sdc_location: a sdc11073.location.SdcLocation instance
        :param validators: a list of InstanceIdentifier objects or None
               If None, self.default_validators is used.
        :param location_context_descriptor_handle: Only needed if the mdib contains more than one
               LocationContextDescriptor. Then this defines the descriptor for which a new LocationContextState
               shall be created.
        """
        mdib = self._mdib
        pm = mdib.data_model.pm_names

        if location_context_descriptor_handle is None:
            # assume there is only one descriptor in mdib, user has not provided a handle.
            location_entities = mdib.entities.by_node_type(pm.LocationContextDescriptor)
            location_entity = None if len(location_entities) == 0 else location_entities[0]
        else:
            location_entity = mdib.entities.by_handle(location_context_descriptor_handle)

        if location_entity is None:
            raise ValueError('no LocationContextDescriptor entity found in mdib')

        new_location = location_entity.new_state()
        new_location.update_from_sdc_location(sdc_location)
        if validators is None:
            new_location.Validator = self.default_validators
        else:
            new_location.Validator = validators

        with mdib.context_state_transaction() as mgr:
            # disassociate before creating a new state
            handles = self.disassociate_all(location_entity,
                                            mgr.new_mdib_version,
                                            ignored_handle=new_location.Handle)
            new_location.BindingMdibVersion = mgr.new_mdib_version
            new_location.BindingStartTime = time.time()
            new_location.ContextAssociation = mdib.data_model.pm_types.ContextAssociation.ASSOCIATED
            handles.append(new_location.Handle)
            mgr.write_entity(location_entity, handles)

    def set_initial_content(self,
                            descriptor_containers: list[AbstractDescriptorContainer],
                            state_containers: list[AbstractStateContainer]):
        """Add states."""
        if self._mdib.is_initialized:  # pragma: no cover
            raise ApiUsageError('method "set_initial_content" can not be called when mdib is already initialized')
        for d in descriptor_containers:
            states = [s for s in state_containers if s.DescriptorHandle == d.Handle]
            entity = self._mdib.entity_factory(d, states)
            self._mdib.internal_entities[d.Handle] = entity

        self.set_all_source_mds()
        self.mk_state_containers_for_all_descriptors()
        self.set_states_initial_values()
        self.update_retrievability_lists()

    def mk_state_containers_for_all_descriptors(self):
        """Create a state container for every descriptor that is missing a state in mdib.

        The model requires that there is a state for every descriptor (exception: multi-states)
        """
        mdib = self._mdib
        pm = mdib.data_model.pm_names
        for entity in mdib.internal_entities.values():
            if entity.descriptor.is_context_descriptor:
                continue
            if entity.state is None:
                state_cls = mdib.data_model.get_state_class_for_descriptor(entity.descriptor)
                state = state_cls(entity.descriptor)
                entity.state = state
                # add some initial values where needed
                if state.is_alert_condition:
                    state.DeterminationTime = time.time()
                elif state.NODETYPE == pm.AlertSystemState:  # noqa: SIM300
                    state.LastSelfCheck = time.time()
                    state.SelfCheckCount = 1
                elif state.NODETYPE == pm.ClockState:  # noqa: SIM300
                    state.LastSet = time.time()
                if mdib.current_transaction is not None:
                    mdib.current_transaction.add_state(state)

    def set_states_initial_values(self):
        """Set all states to defined starting conditions.

        This method is ment to be called directly after the mdib was loaded and before the provider is published
        on the network.
        It changes values only internally in the mdib, no notifications are sent!

        """
        pm_names = self._mdib.data_model.pm_names
        pm_types = self._mdib.data_model.pm_types

        for entity in self._mdib.internal_entities.values():
            if entity.node_type == pm_names.AlertSystemDescriptor:
                # alert systems are active
                entity.state.ActivationState = pm_types.AlertActivation.ON
                entity.state.SystemSignalActivation.append(
                    pm_types.SystemSignalActivation(manifestation=pm_types.AlertSignalManifestation.AUD,
                                                    state=pm_types.AlertActivation.ON))
            elif entity.descriptor.is_alert_condition_descriptor:
                # alert conditions are active, but not present
                entity.state.ActivationState = pm_types.AlertActivation.ON
                entity.state.Presence = False
            elif entity.descriptor.is_alert_signal_descriptor:
                # alert signals are not present, and delegable signals are also not active
                if entity.descriptor.SignalDelegationSupported:
                    entity.state.Location = pm_types.AlertSignalPrimaryLocation.REMOTE
                    entity.state.ActivationState = pm_types.AlertActivation.OFF
                    entity.state.Presence = pm_types.AlertSignalPresence.OFF
                else:
                    entity.state.ActivationState = pm_types.AlertActivation.ON
                    entity.state.Presence = pm_types.AlertSignalPresence.OFF
            elif entity.descriptor.is_component_descriptor:
                # all components are active
                entity.state.ActivationState = pm_types.ComponentActivation.ON
            elif entity.descriptor.is_operational_descriptor:
                # all operations are enabled
                entity.state.OperatingMode = pm_types.OperatingMode.ENABLED

    def update_retrievability_lists(self):
        """Update internal lists, based on current mdib descriptors."""
        mdib = self._mdib
        with mdib.mdib_lock:
            del mdib._retrievability_episodic[:]  # noqa: SLF001
            mdib.retrievability_periodic.clear()
            for entity in mdib.internal_entities.values():
                for r in entity.descriptor.get_retrievability():
                    for r_by in r.By:
                        if r_by.Method == RetrievabilityMethod.EPISODIC:
                            mdib._retrievability_episodic.append(entity.descriptor.Handle)  # noqa: SLF001
                        elif r_by.Method == RetrievabilityMethod.PERIODIC:
                            period_float = r_by.UpdatePeriod
                            period_ms = int(period_float * 1000.0)
                            mdib.retrievability_periodic[period_ms].append(entity.descriptor.Handle)

    def get_all_entities_in_subtree(self, root_entity: ProviderEntity | ProviderMultiStateEntity,
                                    depth_first: bool = True,
                                    include_root: bool = True,
                                    ) -> list[ProviderEntity | ProviderMultiStateEntity]:
        """Return the tree below descriptor_container as a flat list."""
        result = []

        def _getchildren(parent: ProviderEntity | ProviderMultiStateEntity):
            child_containers = [e for e in self._mdib.internal_entities.values() if e.parent_handle == parent.handle]
            if not depth_first:
                result.extend(child_containers)
            apply_map(_getchildren, child_containers)
            if depth_first:
                result.extend(child_containers)

        if include_root and not depth_first:
            result.append(root_entity)
        _getchildren(root_entity)
        if include_root and depth_first:
            result.append(root_entity)
        return result

    def disassociate_all(self,
                         entity: ProviderMultiStateEntity,
                         unbinding_mdib_version: int,
                         ignored_handle: str | None = None) -> list[str]:
        """Disassociate all associated states in entity.

        The method returns a list of states that were disassociated.
        :param entity: ProviderMultiStateEntity
        :param ignored_handle: the context state with this Handle shall not be touched.
        """
        pm_types = self._mdib.data_model.pm_types
        disassociated_state_handles = []
        for state in entity.states.values():
            if state.Handle == ignored_handle or state.ContextAssociation == pm_types.ContextAssociation.NO_ASSOCIATION:
                # If state is already part of this transaction leave it also untouched, accept what the user wanted.
                # If state is not associated, also do not touch it.
                continue
            if state.ContextAssociation != pm_types.ContextAssociation.DISASSOCIATED \
                    or state.UnbindingMdibVersion is None:
                state.ContextAssociation = pm_types.ContextAssociation.DISASSOCIATED
                if state.UnbindingMdibVersion is None:
                    state.UnbindingMdibVersion = unbinding_mdib_version
                    state.BindingEndTime = time.time()
                disassociated_state_handles.append(state.Handle)
        return disassociated_state_handles



from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, ClassVar

from lxml import etree

from sdc11073 import loghelper
from sdc11073.namespaces import PrefixesEnum

if TYPE_CHECKING:
    from sdc11073 import xml_utils
    from sdc11073.dispatch.request import RequestData
    from sdc11073.namespaces import PrefixNamespace
    from sdc11073.pysoap.msgfactory import CreatedMessage
    from sdc11073.xml_types.msg_types import AbstractSet, AbstractSetResponse

msg_prefix = PrefixesEnum.MSG.prefix

WSP_NS = PrefixesEnum.WSP.namespace
WSDL_S12 = PrefixesEnum.WSDL12.namespace  # old soap 12 namespace, used in wsdl 1.1. used only for wsdl

_wsdl_ns = PrefixesEnum.WSDL.namespace
_wsdl_message = etree.QName(_wsdl_ns, 'message')
_wsdl_part = etree.QName(_wsdl_ns, 'part')
_wsdl_operation = etree.QName(_wsdl_ns, 'operation')

# WSDL Generation:
# types to allow declaration of a wsdl data per service
WSDLMessageDescription = namedtuple('WSDLMessageDescription', 'name parameters ')
WSDLOperationBinding = namedtuple('WSDLOperationBinding', 'name input output')


class DPWSPortTypeBase:
    """Base class of all PortType implementations.

    Its responsibilities are:
    - handling of messages
    - creation of wsdl information.
    Handlers are registered in the hosting service instance.
    """

    port_type_name: etree.QName | None = None
    WSDLOperationBindings = ()  # overwrite in derived classes
    WSDLMessageDescriptions = ()  # overwrite in derived classes
    additional_namespaces: ClassVar[list[PrefixNamespace]] = []  # for special namespaces

    def __init__(self, sdc_device, log_prefix: str | None = None):
        """:param sdc_device: the sdc device
        :param log_prefix: optional string
        """
        self._sdc_device = sdc_device
        self._mdib = sdc_device.mdib
        self._sdc_definitions = self._mdib.sdc_definitions
        self._msg_factory = self._sdc_device.msg_factory
        self._data_model = sdc_device.mdib.sdc_definitions.data_model
        self._logger = loghelper.get_logger_adapter(f'sdc.device.{self.__class__.__name__}', log_prefix)
        self.hosting_service = None  # the parent
        # calculate offered subscriptions from WSDLOperationBindings
        self.offered_subscriptions = self._mk_offered_subscriptions()

    def register_hosting_service(self, dpws_hosted_service):
        """Register callbacks in hosting_service."""
        self.hosting_service = dpws_hosted_service

    @property
    def actions(self):  # just a shortcut
        return self._mdib.sdc_definitions.Actions

    def add_wsdl_port_type(self, parent_node):
        raise NotImplementedError

    def _mk_port_type_node(self, parent_node: xml_utils.LxmlElement,
                           is_event_source: bool = False) -> xml_utils.LxmlElement:
        """Needed for wsdl message
        :param parent_node: where to add data
        :param is_event_source: true if port type provides notification
        :return: the new created node (is already child of parent_node).
        """
        if self.port_type_name is None:
            raise ValueError('self.port_type_name is not set, cannot create port type node')

        if 'dt' in parent_node.nsmap:
            port_type = etree.SubElement(parent_node, etree.QName(_wsdl_ns, 'portType'),
                                         attrib={'name': self.port_type_name.localname,
                                                 PrefixesEnum.DPWS.tag('DiscoveryType'): 'dt:ServiceProvider'})
        else:
            port_type = etree.SubElement(parent_node, etree.QName(_wsdl_ns, 'portType'),
                                         attrib={'name': self.port_type_name.localname})
        if is_event_source:
            port_type.attrib[PrefixesEnum.WSE.tag('EventSource')] = 'true'
        return port_type

    def __repr__(self):
        return f'{self.__class__.__name__} Porttype={self.port_type_name!s}'

    def add_wsdl_messages(self, parent_node):
        """Add wsdl:message node to parent_node.
        xml looks like this:
        <wsdl:message name="GetMdDescription">
            <wsdl:part element="msg:GetMdDescription" name="parameters" />
        </wsdl:message>
        :param parent_node:
        :return:
        """
        for msg in self.WSDLMessageDescriptions:
            elem = etree.SubElement(parent_node, _wsdl_message, attrib={'name': msg.name})
            for element_name in msg.parameters:
                etree.SubElement(elem, _wsdl_part,
                                 attrib={'name': 'parameters',
                                         'element': element_name})

    def add_wsdl_binding(self, parent_node, porttype_prefix):
        """Add wsdl:binding node to parent_node.
        xml looks like this:
        <wsdl:binding name="GetBinding" type="msg:Get">
            <s12:binding style="document" transport="http://schemas.xmlsoap.org/soap/http" />
            <wsdl:operation name="GetMdib">
                <s12:operation soapAction="http://p11073-10207/draft6/msg/2016/12/08/Get/GetMdib" />
                <wsdl:input>
                    <s12:body use="literal" />
                </wsdl:input>
                <wsdl:output>
                    <s12:body use="literal" />
                </wsdl:output>
            </wsdl:operation>
            ...
        </wsdl:binding>
        :param parent_node:
        :param porttype_prefix:
        :return:
        """
        v_ref = self._sdc_device.mdib.sdc_definitions
        p_type = self.port_type_name.localname
        wsdl_binding = etree.SubElement(parent_node, etree.QName(_wsdl_ns, 'binding'),
                                        attrib={'name': p_type + 'Binding',
                                                'type': f'{porttype_prefix}:{p_type}'})
        etree.SubElement(wsdl_binding, etree.QName(WSDL_S12, 'binding'),
                         attrib={'style': 'document', 'transport': 'http://schemas.xmlsoap.org/soap/http'})
        _add_policy_dpws_profile(wsdl_binding)
        for wsdl_op in self.WSDLOperationBindings:
            wsdl_operation = etree.SubElement(wsdl_binding, etree.QName(_wsdl_ns, 'operation'),
                                              attrib={'name': wsdl_op.name})
            etree.SubElement(wsdl_operation, etree.QName(WSDL_S12, 'operation'),
                             attrib={'soapAction': f'{v_ref.ActionsNamespace}/{p_type}/{wsdl_op.name}'})
            if wsdl_op.input is not None:
                wsdl_input = etree.SubElement(wsdl_operation, etree.QName(_wsdl_ns, 'input'))
                etree.SubElement(wsdl_input, etree.QName(WSDL_S12, 'body'), attrib={'use': wsdl_op.input})
            if wsdl_op.output is not None:
                wsdl_output = etree.SubElement(wsdl_operation, etree.QName(_wsdl_ns, 'output'))
                etree.SubElement(wsdl_output, etree.QName(WSDL_S12, 'body'), attrib={'use': wsdl_op.output})

    def _mk_offered_subscriptions(self) -> list:
        """Takes action strings from sdc_definitions.Actions.
        The name of the WSDLOperationBinding is used to reference the action string.
        """
        actions = self._sdc_device.mdib.sdc_definitions.Actions
        offered_subscriptions = []
        for bdg in self.WSDLOperationBindings:
            if bdg.input is None:
                action_string = getattr(actions, bdg.name)
                offered_subscriptions.append(action_string)
        return offered_subscriptions


class ServiceWithOperations(DPWSPortTypeBase):

    def _handle_operation_request(self, request_data: RequestData,
                                  request: AbstractSet,
                                  set_response: AbstractSetResponse) -> CreatedMessage:
        """Handle thew operation request by forwarding it to provider."""
        data_model = self._sdc_definitions.data_model
        operation = self._sdc_device.get_operation_by_handle(request.OperationHandleRef)
        transaction_id = self._sdc_device.generate_transaction_id()
        set_response.InvocationInfo.TransactionId = transaction_id
        if operation is None:
            error_text = f'no handler registered for "{request.OperationHandleRef}"'
            self._logger.warning('handle operation request: {}', error_text)
            set_response.InvocationInfo.InvocationState = data_model.msg_types.InvocationState.FAILED
            set_response.InvocationInfo.InvocationError = data_model.msg_types.InvocationError.INVALID_VALUE
            set_response.InvocationInfo.add_error_message(error_text)
        else:
            invocation_state = self._sdc_device.handle_operation_request(operation,
                                                                         request_data.message_data.p_msg,
                                                                         request,
                                                                         transaction_id)
            self._logger.info('operation request "{}" handled, transaction id = {}, invocation-state={}',
                              request.OperationHandleRef, set_response.InvocationInfo.TransactionId, invocation_state)
            set_response.InvocationInfo.InvocationState = invocation_state

        set_response.MdibVersion = self._mdib.mdib_version
        set_response.SequenceId = self._mdib.sequence_id
        set_response.InstanceId = self._mdib.instance_id
        return self._sdc_device.msg_factory.mk_reply_soap_message(request_data, set_response)


def _mk_wsdl_operation(parent_node, operation_name, input_message_name, output_message_name) -> xml_utils.LxmlElement:
    elem = etree.SubElement(parent_node, _wsdl_operation, attrib={'name': operation_name})
    if input_message_name is not None:
        etree.SubElement(elem, etree.QName(_wsdl_ns, 'input'),
                         attrib={'message': f'tns:{input_message_name}',
                                 })
    if output_message_name is not None:
        etree.SubElement(elem, etree.QName(_wsdl_ns, 'output'),
                         attrib={'message': f'tns:{output_message_name}',
                                 })
    return elem


def mk_wsdl_two_way_operation(parent_node: xml_utils.LxmlElement,
                              operation_name: str,
                              input_message_name: str | None = None,
                              output_message_name: str | None = None) -> xml_utils.LxmlElement:
    """A helper for wsdl generation. A two-way-operation defines a 'normal' request and response operation.
    :param parent_node: info shall be added to this node
    :param operation_name: a string
    :param input_message_name: only needed if message name is not equal to operation_name
    :param output_message_name: only needed if message name is not equal to operation_name + "Response"
    :return:
    """
    input_msg_name = input_message_name or operation_name  # defaults to operation name
    output_msg_name = output_message_name or operation_name + 'Response'  # defaults to operation name + "Response"
    return _mk_wsdl_operation(parent_node,
                              operation_name=operation_name,
                              input_message_name=input_msg_name,
                              output_message_name=output_msg_name)


def mk_wsdl_one_way_operation(parent_node: xml_utils.LxmlElement,
                              operation_name: str,
                              output_message_name: str | None = None) -> xml_utils.LxmlElement:
    """A helper for wsdl generation. A one-way-operation is a subscription.
    :param parent_node: info shall be added to this node
    :param operation_name: a string
    :param output_message_name: only needed if message name is not equal to operation_name
    :return:
    """
    output_msg_name = output_message_name or operation_name  # defaults to operation name
    return _mk_wsdl_operation(parent_node,
                              operation_name=operation_name,
                              input_message_name=None,
                              output_message_name=output_msg_name)


def _add_policy_dpws_profile(parent_node):
    """:param parent_node:
    :return: <wsp:Policy>
            <dpws:Profile wsp:Optional="true"/>
            <mdpws:Profile wsp:Optional="true"/>
          </wsp:Policy>
    """
    wsp_policy_node = etree.SubElement(parent_node, etree.QName(WSP_NS, 'Policy'), attrib=None)
    _ = etree.SubElement(wsp_policy_node, PrefixesEnum.DPWS.tag('Profile'),
                         attrib={etree.QName(WSP_NS, 'Optional'): 'true'})
    _ = etree.SubElement(wsp_policy_node, PrefixesEnum.MDPWS.tag('Profile'),
                         attrib={etree.QName(WSP_NS, 'Optional'): 'true'})



"""The module contains a provider mdib implementation that uses entities in internal representation."""
from __future__ import annotations

import uuid
from collections import defaultdict
from contextlib import AbstractContextManager, contextmanager
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable, Protocol

from lxml.etree import Element, SubElement

from sdc11073 import loghelper
from sdc11073.definitions_base import ProtocolsRegistry
from sdc11073.mdib.transactionsprotocol import TransactionType
from sdc11073.observableproperties import ObservableProperty
from sdc11073.pysoap.msgreader import MessageReader

from .entities import ProviderInternalEntity, ProviderInternalEntityType, ProviderInternalMultiStateEntity
from .entity_mdibbase import EntityMdibBase
from .entity_providermdibxtra import EntityProviderMdibMethods
from .entity_transactions import mk_transaction

if TYPE_CHECKING:
    from collections.abc import Iterable

    from lxml.etree import QName

    from sdc11073 import xml_utils
    from sdc11073.definitions_base import BaseDefinitions
    from sdc11073.loghelper import LoggerAdapter
    from sdc11073.mdib.descriptorcontainers import AbstractDescriptorContainer
    from sdc11073.mdib.entityprotocol import ProviderEntityGetterProtocol
    from sdc11073.mdib.mdibbase import MdibVersionGroup
    from sdc11073.mdib.statecontainers import AbstractStateContainer
    from sdc11073.mdib.transactionsprotocol import (
        AnyEntityTransactionManagerProtocol,
        EntityContextStateTransactionManagerProtocol,
        EntityDescriptorTransactionManagerProtocol,
        EntityStateTransactionManagerProtocol,
        TransactionResultProtocol,
    )
    from sdc11073.xml_types.pm_types import CodedValue, Coding

    from .entities import ProviderEntityType, ProviderMultiStateEntity

    ProviderEntityFactory = Callable[[AbstractDescriptorContainer, list[AbstractStateContainer]],
    ProviderInternalEntityType]


def _mk_internal_entity(descriptor_container: AbstractDescriptorContainer,
                        states: list[AbstractStateContainer]) -> ProviderInternalEntityType:
    """Create an entity.

    This is the default Implementation of ProviderEntityFactory.
    """
    for s in states:
        s.descriptor_container = descriptor_container
    if descriptor_container.is_context_descriptor:
        return ProviderInternalMultiStateEntity(descriptor_container, states)
    if len(states) == 1:
        return ProviderInternalEntity(descriptor_container, states[0])
    if len(states) == 0:
        return ProviderInternalEntity(descriptor_container, None)
    msg = f'found {len(states)} states for {descriptor_container.NODETYPE} handle = {descriptor_container.Handle}'
    raise ValueError(msg)


class ProviderEntityGetter:
    """Implements entityprotocol.ProviderEntityGetterProtocol."""

    def __init__(self,
                 mdib: EntityProviderMdib):
        self._mdib = mdib

    def by_handle(self, handle: str) -> ProviderEntityType | None:
        """Return entity with given handle."""
        with self._mdib.mdib_lock:
            try:
                internal_entity = self._mdib.internal_entities[handle]
                return internal_entity.mk_entity(self._mdib)
            except KeyError:
                return None

    def by_context_handle(self, handle: str) -> ProviderMultiStateEntity | None:
        """Return multi state entity that contains a state with given handle."""
        with self._mdib.mdib_lock:
            for internal_entity in self._mdib.internal_entities.values():
                if internal_entity.is_multi_state and handle in internal_entity.states:
                    return internal_entity.mk_entity(self._mdib)
        return None

    def by_node_type(self, node_type: QName) -> list[ProviderEntityType]:
        """Return all entities with given node type."""
        ret = []
        with self._mdib.mdib_lock:
            for internal_entity in self._mdib.internal_entities.values():
                if node_type == internal_entity.descriptor.NODETYPE:
                    ret.append(internal_entity.mk_entity(self._mdib)) # noqa: PERF401
        return ret

    def by_parent_handle(self, parent_handle: str | None) -> list[ProviderEntityType]:
        """Return all entities with given parent handle."""
        ret = []
        with self._mdib.mdib_lock:
            for internal_entity in self._mdib.internal_entities.values():
                if internal_entity.descriptor.parent_handle == parent_handle:
                    ret.append(internal_entity.mk_entity(self._mdib))   # noqa: PERF401
        return ret

    def by_coding(self, coding: Coding) -> list[ProviderEntityType]:
        """Return all entities with given Coding."""
        ret = []
        with (self._mdib.mdib_lock):
            for internal_entity in self._mdib.internal_entities.values():
                if internal_entity.descriptor.Type is not None \
                    and internal_entity.descriptor.Type.is_equivalent(coding):
                    ret.append(internal_entity.mk_entity(self._mdib))  # noqa: PERF401
        return ret

    def by_coded_value(self, coded_value: CodedValue) -> list[ProviderEntityType]:
        """Return all entities with given Coding."""
        ret = []
        with self._mdib.mdib_lock:
            for internal_entity in self._mdib.internal_entities.values():
                if internal_entity.descriptor.Type is not None and internal_entity.descriptor.Type.is_equivalent(
                        coded_value):
                    ret.append(internal_entity.mk_entity(self._mdib))  # noqa: PERF401
        return ret

    def items(self) -> Iterable[tuple[str, ProviderEntityType]]:
        """Return the items."""
        with self._mdib.mdib_lock:
            for handle, internal_entity in self._mdib.internal_entities.items():
                yield handle, internal_entity.mk_entity(self._mdib)

    def new_entity(self,
                   node_type: QName,
                   handle: str,
                   parent_handle: str | None) -> ProviderEntityType:
        """Create an entity.

        User can modify the entity and then add it to transaction via write_entity!
        It will not become part of mdib without write_entity call!
        """
        if handle in self._mdib.internal_entities or handle in self._mdib.new_entities:
            raise ValueError('Handle already exists')

        descr_cls = self._mdib.data_model.get_descriptor_container_class(node_type)
        descriptor_container = descr_cls(handle=handle, parent_handle=parent_handle)
        if parent_handle is not None:
            parent_entity = (self._mdib.new_entities.get(parent_handle)
                             or self._mdib.internal_entities.get(parent_handle))
            if parent_entity is None: # pragma: no cover
                msg = f'Entity {handle} has no parent (parent_handle = {parent_handle})!'
                raise ValueError(msg)
            descriptor_container.set_source_mds(parent_entity.descriptor.source_mds)
        else:
            descriptor_container.set_source_mds(descriptor_container.Handle)  # this is a mds, source_mds is itself

        new_internal_entity = self._mdib.entity_factory(descriptor_container, [])
        if handle in self._mdib.descr_handle_version_lookup:
            # This handle existed before. Use last descriptor version + 1
            new_internal_entity.descriptor.DescriptorVersion = self._mdib.descr_handle_version_lookup[handle] + 1
        if not new_internal_entity.is_multi_state:
            # create a state
            state_cls = self._mdib.data_model.get_state_container_class(descriptor_container.STATE_QNAME)
            new_internal_entity.state = state_cls(descriptor_container)
            if handle in self._mdib.state_handle_version_lookup:
                new_internal_entity.state.StateVersion = self._mdib.state_handle_version_lookup[handle] + 1
        # write to mdib in process_transaction
        self._mdib.new_entities[descriptor_container.Handle] = new_internal_entity
        return new_internal_entity.mk_entity(self._mdib)

    def __len__(self) -> int:
        """Return number of entities."""
        return len(self._mdib.internal_entities)


class ProviderXtraProtocol(Protocol):
    """Functionality expected by EntityProviderMdib."""

    def set_initial_content(self,
                            descriptor_containers: list[AbstractDescriptorContainer],
                            state_containers: list[AbstractStateContainer]):
        """Populate mdib."""
        ...


class EntityProviderMdib(EntityMdibBase):
    """Device side implementation of a mdib.

    Do not modify containers directly, use transactions for that purpose.
    Transactions keep track of changes and initiate sending of update notifications to clients.
    """

    transaction: TransactionResultProtocol | None = ObservableProperty(fire_only_on_changed_value=False)
    rt_updates = ObservableProperty(fire_only_on_changed_value=False)  # different observable for performance

    def __init__(self,
                 sdc_definitions: type[BaseDefinitions] | None = None,
                 log_prefix: str | None = None,
                 extra_functionality: Callable[[EntityProviderMdib], ProviderXtraProtocol] | None = None,
                 transaction_factory: Callable[[EntityProviderMdib, TransactionType, LoggerAdapter],
                                                AnyEntityTransactionManagerProtocol] | None = None,
                 ):
        """Construct a ProviderMdib.

        :param sdc_definitions: defaults to sdc11073.definitions_sdc.SdcV1Definitions
        :param log_prefix: a string
        :param extra_functionality: class for extra functionality, default is ProviderMdibMethods
        :param transaction_factory: optional alternative transactions factory.
        """
        if sdc_definitions is None:  # pragma: no cover
            from sdc11073.definitions_sdc import SdcV1Definitions  # lazy import, needed to brake cyclic imports
            sdc_definitions = SdcV1Definitions
        super().__init__(sdc_definitions,
                         loghelper.get_logger_adapter('sdc.device.mdib', log_prefix),
                         )

        self.nsmapper = sdc_definitions.data_model.ns_helper

        if extra_functionality is None:
            extra_functionality = EntityProviderMdibMethods

        self._entities: dict[str, ProviderInternalEntityType] = {}  # key is the handle

        # Keep track of entities that were created but are not yet part of mdib.
        # They become part of mdib when they are added via transaction.
        self._new_entities: dict[str, ProviderInternalEntityType] = {}

        # The official API
        self.entities: ProviderEntityGetterProtocol = ProviderEntityGetter(self)

        self._xtra = extra_functionality(self)
        self._tr_lock = Lock()  # transaction lock

        self.sequence_id = uuid.uuid4().urn  # this uuid identifies this mdib instance

        self._annotators = {}
        self.current_transaction = None

        self.pre_commit_handler = None  # pre_commit_handler can modify transaction if needed before it is committed
        self.post_commit_handler = None  # post_commit_handler can modify mdib if needed after it is committed
        self._transaction_factory = transaction_factory or mk_transaction
        self._retrievability_episodic = []  # a list of handles
        self.retrievability_periodic = defaultdict(list)
        self.mddescription_version = 0
        self.mdstate_version = 0
        self._is_initialized = False
        # In order to be able to re-create a descriptor or state with a bigger version than before,
        # these lookups keep track of version counters for deleted descriptors and states.
        self.descr_handle_version_lookup: dict[str, int] = {}
        self.state_handle_version_lookup: dict[str, int] = {}
        self.entity_factory: ProviderEntityFactory = _mk_internal_entity

    @property
    def xtra(self) -> Any:
        """Give access to extended functionality."""
        return self._xtra

    @property
    def internal_entities(self) -> dict[str, ProviderInternalEntityType]:
        """The property is needed by transactions. Do not use it otherwise."""
        return self._entities

    @property
    def new_entities(self) -> dict[str, ProviderInternalEntityType]:
        """The property is needed by transactions. Do not use it otherwise."""
        return self._new_entities

    def set_initialized(self):
        """Set initialized state = True."""
        self._is_initialized = True

    @property
    def is_initialized(self) -> bool:
        """Return True if mdib is already initialized."""
        return self._is_initialized

    @contextmanager
    def _transaction_manager(self, # noqa: PLR0912, C901
                             transaction_type: TransactionType,
                             set_determination_time: bool = True) -> AbstractContextManager[
        AnyEntityTransactionManagerProtocol]:
        """Start a transaction, return a new transaction manager."""
        with self._tr_lock, self.mdib_lock:
            try:
                self.current_transaction = self._transaction_factory(self, transaction_type, self.logger)
                yield self.current_transaction

                if callable(self.pre_commit_handler):
                    self.pre_commit_handler(self, self.current_transaction)
                if self.current_transaction.error:
                    self._logger.info('transaction_manager: transaction without updates!')
                else:
                    # update observables
                    transaction_result = self.current_transaction.process_transaction(set_determination_time)
                    if transaction_result.new_mdib_version is not None:
                        self.mdib_version = transaction_result.new_mdib_version
                    self.transaction = transaction_result

                    if transaction_result.alert_updates:
                        self.alert_by_handle = {st.DescriptorHandle: st for st in transaction_result.alert_updates}
                    if transaction_result.comp_updates:
                        self.component_by_handle = {st.DescriptorHandle: st for st in transaction_result.comp_updates}
                    if transaction_result.ctxt_updates:
                        self.context_by_handle = {st.Handle: st for st in transaction_result.ctxt_updates}
                    if transaction_result.descr_created:
                        self.new_descriptors_by_handle = {descr.Handle: descr for descr
                                                          in transaction_result.descr_created}
                    if transaction_result.descr_deleted:
                        self.deleted_descriptors_by_handle = {descr.Handle: descr for descr
                                                              in transaction_result.descr_deleted}
                    if transaction_result.descr_updated:
                        self.updated_descriptors_by_handle = {descr.Handle: descr for descr
                                                              in transaction_result.descr_updated}
                    if transaction_result.metric_updates:
                        self.metrics_by_handle = {st.DescriptorHandle: st for st in transaction_result.metric_updates}
                    if transaction_result.op_updates:
                        self.operation_by_handle = {st.DescriptorHandle: st for st in transaction_result.op_updates}
                    if transaction_result.rt_updates:
                        self.waveform_by_handle = {st.DescriptorHandle: st for st in transaction_result.rt_updates}

                    if callable(self.post_commit_handler):
                        self.post_commit_handler(self, self.current_transaction)
            finally:
                self.current_transaction = None

    @contextmanager
    def context_state_transaction(self) -> AbstractContextManager[EntityContextStateTransactionManagerProtocol]:
        """Return a transaction for context state updates."""
        with self._transaction_manager(TransactionType.context, False) as mgr:
            yield mgr

    @contextmanager
    def alert_state_transaction(self, set_determination_time: bool = True) \
            -> AbstractContextManager[EntityStateTransactionManagerProtocol]:
        """Return a transaction for alert state updates."""
        with self._transaction_manager(TransactionType.alert, set_determination_time) as mgr:
            yield mgr

    @contextmanager
    def metric_state_transaction(self, set_determination_time: bool = True) \
            -> AbstractContextManager[EntityStateTransactionManagerProtocol]:
        """Return a transaction for metric state updates (not real time samples!)."""
        with self._transaction_manager(TransactionType.metric, set_determination_time) as mgr:
            yield mgr

    @contextmanager
    def rt_sample_state_transaction(self, set_determination_time: bool = False) \
            -> AbstractContextManager[EntityStateTransactionManagerProtocol]:
        """Return a transaction for real time sample state updates."""
        with self._transaction_manager(TransactionType.rt_sample, set_determination_time) as mgr:
            yield mgr

    @contextmanager
    def component_state_transaction(self) -> AbstractContextManager[EntityStateTransactionManagerProtocol]:
        """Return a transaction for component state updates."""
        with self._transaction_manager(TransactionType.component) as mgr:
            yield mgr

    @contextmanager
    def operational_state_transaction(self) -> AbstractContextManager[EntityStateTransactionManagerProtocol]:
        """Return a transaction for operational state updates."""
        with self._transaction_manager(TransactionType.operational) as mgr:
            yield mgr

    @contextmanager
    def descriptor_transaction(self) -> AbstractContextManager[EntityDescriptorTransactionManagerProtocol]:
        """Return a transaction for descriptor updates.

        This transaction also allows to handle the states that relate to the modified descriptors.
        """
        with self._transaction_manager(TransactionType.descriptor) as mgr:
            yield mgr

    def make_descriptor_node(self,
                             descriptor_container: AbstractDescriptorContainer,
                             parent_node: xml_utils.LxmlElement,
                             tag: QName,
                             set_xsi_type: bool = True) -> xml_utils.LxmlElement:
        """Create a lxml etree node with subtree from instance data.

        :param descriptor_container: a descriptor container instance
        :param parent_node: parent node
        :param tag: tag of node
        :param set_xsi_type: if true, the NODETYPE will be used to set the xsi:type attribute of the node
        :return: an etree node.
        """
        ns_map = self.nsmapper.partial_map(self.nsmapper.PM, self.nsmapper.XSI) \
            if set_xsi_type else self.nsmapper.partial_map(self.nsmapper.PM)
        node = SubElement(parent_node,
                          tag,
                          attrib={'Handle': descriptor_container.Handle},
                          nsmap=ns_map)
        descriptor_container.update_node(node, self.nsmapper, set_xsi_type)  # create all
        child_entities = self.entities.by_parent_handle(descriptor_container.Handle)
        # append all child containers, then bring all child elements in correct order
        for child_entity in child_entities:
            child_tag, set_xsi = descriptor_container.tag_name_for_child_descriptor(child_entity.descriptor.NODETYPE)
            self.make_descriptor_node(child_entity.descriptor, node, child_tag, set_xsi)
        descriptor_container.sort_child_nodes(node)
        return node

    def reconstruct_mdib(self) -> tuple[xml_utils.LxmlElement, MdibVersionGroup]:
        """Build dom tree from current data.

        This method does not include context states!
        """
        with self.mdib_lock:
            return self._reconstruct_mdib(add_context_states=False), self.mdib_version_group

    def reconstruct_mdib_with_context_states(self) -> tuple[xml_utils.LxmlElement, MdibVersionGroup]:
        """Build dom tree from current data.

        This method includes the context states.
        """
        with self.mdib_lock:
            return self._reconstruct_mdib(add_context_states=True), self.mdib_version_group

    def reconstruct_md_description(self) -> tuple[xml_utils.LxmlElement, MdibVersionGroup]:
        """Build dom tree of descriptors from current data."""
        with self.mdib_lock:
            node = self._reconstruct_md_description()
            return node, self.mdib_version_group

    def _reconstruct_mdib(self, add_context_states: bool) -> xml_utils.LxmlElement:
        """Build dom tree of mdib from current data.

        If add_context_states is False, context states are not included.
        """
        pm = self.data_model.pm_names
        msg = self.data_model.msg_names
        doc_nsmap = self.nsmapper.ns_map
        mdib_node = Element(msg.Mdib, nsmap=doc_nsmap)
        mdib_node.set('MdibVersion', str(self.mdib_version))
        mdib_node.set('SequenceId', self.sequence_id)
        if self.instance_id is not None:
            mdib_node.set('InstanceId', str(self.instance_id))
        md_description_node = self._reconstruct_md_description()
        mdib_node.append(md_description_node)

        # add a list of states
        md_state_node = SubElement(mdib_node, pm.MdState,
                                   attrib={'StateVersion': str(self.mdstate_version)},
                                   nsmap=doc_nsmap)
        tag = pm.State
        for entity in self._entities.values():
            if entity.descriptor.is_context_descriptor:
                if add_context_states:
                    for state_container in entity.states.values():
                        md_state_node.append(state_container.mk_state_node(tag, self.nsmapper))
            elif entity.state is not None:
                md_state_node.append(entity.state.mk_state_node(tag, self.nsmapper))
        return mdib_node

    def _reconstruct_md_description(self) -> xml_utils.LxmlElement:
        """Build dom tree of descriptors from current data."""
        pm = self.data_model.pm_names
        doc_nsmap = self.nsmapper.ns_map
        root_entities = self.entities.by_parent_handle(None)
        md_description_node = Element(pm.MdDescription,
                                      attrib={'DescriptionVersion': str(self.mddescription_version)},
                                      nsmap=doc_nsmap)
        for root_entity in root_entities:
            self.make_descriptor_node(root_entity.descriptor, md_description_node, tag=pm.Mds, set_xsi_type=False)
        return md_description_node

    @classmethod
    def from_mdib_file(cls,
                       path: str,
                       protocol_definition: type[BaseDefinitions] | None = None,
                       xml_reader_class: type[MessageReader] | None = MessageReader,
                       log_prefix: str | None = None) -> EntityProviderMdib:
        """Construct mdib from a file.

        :param path: the input file path for creating the mdib
        :param protocol_definition: an optional object derived from BaseDefinitions, forces usage of this definition
        :param xml_reader_class: class that is used to read mdib xml file
        :param log_prefix: a string or None
        :return: instance.
        """
        with Path(path).open('rb') as the_file:
            xml_text = the_file.read()
        return cls.from_string(xml_text,
                               protocol_definition,
                               xml_reader_class,
                               log_prefix)

    @classmethod
    def from_string(cls,
                    xml_text: bytes,
                    protocol_definition: type[BaseDefinitions] | None = None,
                    xml_reader_class: type[MessageReader] = MessageReader,
                    log_prefix: str | None = None) -> EntityProviderMdib:
        """Construct mdib from a string.

        :param xml_text: the input string for creating the mdib
        :param protocol_definition: an optional object derived from BaseDefinitions, forces usage of this definition
        :param xml_reader_class: class that is used to read mdib xml file
        :param log_prefix: a string or None
        :return: instance.
        """
        # get protocol definition that matches xml_text
        if protocol_definition is None:
            for definition_cls in ProtocolsRegistry.protocols:
                pm_namespace = definition_cls.data_model.ns_helper.PM.namespace.encode('utf-8')
                if pm_namespace in xml_text:
                    protocol_definition = definition_cls
                    break
        if protocol_definition is None:
            raise ValueError('cannot create instance, no known BICEPS schema version identified')
        mdib = cls(protocol_definition, log_prefix=log_prefix)

        xml_msg_reader = xml_reader_class(protocol_definition, None, mdib.logger)
        descriptor_containers, state_containers = xml_msg_reader.read_mdib_xml(xml_text)
        mdib.xtra.set_initial_content(descriptor_containers, state_containers)
        mdib.set_initialized()
        return mdib



from typing import Union, Any
from sdc11073.pysoap.soapenvelope import Fault, faultcodeEnum
from sdc11073.exceptions import ApiUsageError
from sdc11073.exceptions import InvalidPathError

class PathElementRegistry:
    """ A string - object lookup"""

    def __init__(self):
        self._instances = {}

    def register_instance(self, path_element: Union[str, None], instance: Any):
        if path_element in self._instances:
            raise ApiUsageError(f'Path-element "{path_element}" already registered')
        self._instances[path_element] = instance

    def get_instance(self, path_element: Union[str, None]) -> Any:
        instance = self._instances.get(path_element)
        if instance is None:
            fault = Fault()
            fault.Code.Value = faultcodeEnum.SENDER
            fault.add_reason_text(f'invalid path {path_element}')

            raise InvalidPathError(reason=f'{path_element} not found', soap_fault=fault)
        return instance
